import numpy as np
import openmdao.api as om
import wisdem.commonse.utilities as util
import wisdem.pyframe3dd.pyframe3dd as pyframe3dd
import wisdem.commonse.utilization_dnvgl as util_dnvgl
import wisdem.commonse.utilization_constraints as util_con
from wisdem.commonse import NFREQ, gravity
from wisdem.floatingse.member import NULL, MEMMAX, Member

NNODES_MAX = 1000
NELEM_MAX = 1000
RIGID = 1e30
EPS = 1e-6


class PlatformFrame(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("options")

    def setup(self):
        opt = self.options["options"]
        n_member = opt["floating"]["members"]["n_members"]

        for k in range(n_member):
            self.add_input(f"member{k}:nodes_xyz", NULL * np.ones((MEMMAX, 3)), units="m")
            self.add_input(f"member{k}:nodes_r", NULL * np.ones(MEMMAX), units="m")
            self.add_input(f"member{k}:section_D", NULL * np.ones(MEMMAX), units="m")
            self.add_input(f"member{k}:section_t", NULL * np.ones(MEMMAX), units="m")
            self.add_input(f"member{k}:section_A", NULL * np.ones(MEMMAX), units="m**2")
            self.add_input(f"member{k}:section_Asx", NULL * np.ones(MEMMAX), units="m**2")
            self.add_input(f"member{k}:section_Asy", NULL * np.ones(MEMMAX), units="m**2")
            self.add_input(f"member{k}:section_Ixx", NULL * np.ones(MEMMAX), units="kg*m**2")
            self.add_input(f"member{k}:section_Iyy", NULL * np.ones(MEMMAX), units="kg*m**2")
            self.add_input(f"member{k}:section_Izz", NULL * np.ones(MEMMAX), units="kg*m**2")
            self.add_input(f"member{k}:section_rho", NULL * np.ones(MEMMAX), units="kg/m**3")
            self.add_input(f"member{k}:section_E", NULL * np.ones(MEMMAX), units="Pa")
            self.add_input(f"member{k}:section_G", NULL * np.ones(MEMMAX), units="Pa")
            self.add_input(f"member{k}:section_sigma_y", NULL * np.ones(MEMMAX), units="Pa")
            self.add_input(f"member{k}:idx_cb", 0)
            self.add_input(f"member{k}:buoyancy_force", 0.0, units="N")
            self.add_input(f"member{k}:displacement", 0.0, units="m**3")
            self.add_input(f"member{k}:center_of_buoyancy", np.zeros(3), units="m")
            self.add_input(f"member{k}:center_of_mass", np.zeros(3), units="m")
            self.add_input(f"member{k}:ballast_mass", 0.0, units="kg")
            self.add_input(f"member{k}:total_mass", 0.0, units="kg")
            self.add_input(f"member{k}:total_cost", 0.0, units="USD")
            self.add_input(f"member{k}:I_total", np.zeros(6), units="kg*m**2")
            self.add_input(f"member{k}:Awater", 0.0, units="m**2")
            self.add_input(f"member{k}:Iwater", 0.0, units="m**4")
            self.add_input(f"member{k}:added_mass", np.zeros(6), units="kg")
            self.add_input(f"member{k}:waterline_centroid", np.zeros(2), units="m")
            self.add_input(f"member{k}:variable_ballast_capacity", val=0.0, units="m**3")
            self.add_input(f"member{k}:Px", np.zeros(MEMMAX), units="N/m")
            self.add_input(f"member{k}:Py", np.zeros(MEMMAX), units="N/m")
            self.add_input(f"member{k}:Pz", np.zeros(MEMMAX), units="N/m")
            self.add_input(f"member{k}:qdyn", np.zeros(MEMMAX), units="Pa")

        self.add_input("transition_node", np.zeros(3), units="m")
        self.add_input("transition_piece_mass", 0.0, units="kg")
        self.add_input("transition_piece_cost", 0.0, units="USD")

        self.add_output("transition_piece_I", np.zeros(6), units="kg*m**2")

        self.add_output("platform_nodes", NULL * np.ones((NNODES_MAX, 3)), units="m")
        self.add_output("platform_Fnode", NULL * np.ones((NNODES_MAX, 3)), units="N")
        self.add_output("platform_Rnode", NULL * np.ones(NNODES_MAX), units="m")
        self.add_output("platform_elem_n1", NULL * np.ones(NELEM_MAX, dtype=np.int_))
        self.add_output("platform_elem_n2", NULL * np.ones(NELEM_MAX, dtype=np.int_))
        self.add_output("platform_elem_D", NULL * np.ones(NELEM_MAX), units="m")
        self.add_output("platform_elem_t", NULL * np.ones(NELEM_MAX), units="m")
        self.add_output("platform_elem_A", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_output("platform_elem_Asx", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_output("platform_elem_Asy", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_output("platform_elem_Ixx", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_output("platform_elem_Iyy", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_output("platform_elem_Izz", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_output("platform_elem_rho", NULL * np.ones(NELEM_MAX), units="kg/m**3")
        self.add_output("platform_elem_E", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_output("platform_elem_G", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_output("platform_elem_sigma_y", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_output("platform_elem_Px1", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_output("platform_elem_Px2", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_output("platform_elem_Py1", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_output("platform_elem_Py2", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_output("platform_elem_Pz1", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_output("platform_elem_Pz2", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_output("platform_elem_qdyn", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_discrete_output("platform_elem_memid", [-1] * NELEM_MAX)
        self.add_output("platform_displacement", 0.0, units="m**3")
        self.add_output("platform_center_of_buoyancy", np.zeros(3), units="m")
        self.add_output("platform_hull_center_of_mass", np.zeros(3), units="m")
        self.add_output("platform_centroid", np.zeros(3), units="m")
        self.add_output("platform_ballast_mass", 0.0, units="kg")
        self.add_output("platform_hull_mass", 0.0, units="kg")
        self.add_output("platform_mass", 0.0, units="kg")
        self.add_output("platform_I_hull", np.zeros(6), units="kg*m**2")
        self.add_output("platform_cost", 0.0, units="USD")
        self.add_output("platform_Awater", 0.0, units="m**2")
        self.add_output("platform_Iwater", 0.0, units="m**4")
        self.add_output("platform_added_mass", np.zeros(6), units="kg")
        self.add_output("platform_variable_capacity", np.zeros(n_member), units="m**3")

        self.node_mem2glob = {}

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Seems like we have to run this each time as numbering can change during optimization
        self.node_mem2glob = {}
        self.set_connectivity(inputs, outputs)

        self.set_node_props(inputs, outputs)
        self.set_element_props(inputs, outputs, discrete_inputs, discrete_outputs)

    def set_connectivity(self, inputs, outputs):
        # Load in number of members
        opt = self.options["options"]
        n_member = opt["floating"]["members"]["n_members"]

        # Initialize running lists across all members
        nodes_temp = np.empty((0, 3))
        elem_n1 = np.array([], dtype=np.int_)
        elem_n2 = np.array([], dtype=np.int_)

        # Look over members and grab all nodes and internal connections
        for k in range(n_member):
            inode_xyz = inputs[f"member{k}:nodes_xyz"]
            inodes = np.where(inode_xyz[:, 0] == NULL)[0][0]
            inode_xyz = inode_xyz[:inodes, :]
            inode_range = np.arange(inodes - 1)

            n = nodes_temp.shape[0]
            for ii in range(inodes):
                self.node_mem2glob[(k, ii)] = n + ii

            elem_n1 = np.append(elem_n1, n + inode_range)
            elem_n2 = np.append(elem_n2, n + inode_range + 1)
            nodes_temp = np.append(nodes_temp, inode_xyz, axis=0)

        # Reveal connectivity by using mapping to unique node positions
        nodes, idx, inv = np.unique(nodes_temp.round(8), axis=0, return_index=True, return_inverse=True)
        nnode = nodes.shape[0]
        outputs["platform_nodes"] = NULL * np.ones((NNODES_MAX, 3))
        outputs["platform_nodes"][:nnode, :] = nodes
        outputs["platform_centroid"] = nodes.mean(axis=0)

        # Use mapping to set references to node joints
        nelem = elem_n1.size
        outputs["platform_elem_n1"] = NULL * np.ones(NELEM_MAX, dtype=np.int_)
        outputs["platform_elem_n2"] = NULL * np.ones(NELEM_MAX, dtype=np.int_)
        outputs["platform_elem_n1"][:nelem] = inv[elem_n1]
        outputs["platform_elem_n2"][:nelem] = inv[elem_n2]

        # Update global 2 member mappings
        for k in self.node_mem2glob.keys():
            self.node_mem2glob[k] = inv[self.node_mem2glob[k]]

    def set_node_props(self, inputs, outputs):
        # Load in number of members
        opt = self.options["options"]
        n_member = opt["floating"]["members"]["n_members"]

        # Number of valid nodes
        node_platform = outputs["platform_nodes"]
        nnode = np.where(node_platform[:, 0] == NULL)[0][0]
        node_platform = node_platform[:nnode, :]

        # Find greatest radius of all members at node intersections
        Rnode = np.zeros(nnode)
        for k in range(n_member):
            irnode = inputs[f"member{k}:nodes_r"]
            n = np.where(irnode == NULL)[0][0]
            for ii in range(n):
                iglob = self.node_mem2glob[(k, ii)]
                Rnode[iglob] = np.array([Rnode[iglob], irnode[ii]]).max()

        # Find forces on nodes
        Fnode = np.zeros((nnode, 3))
        for k in range(n_member):
            icb = int(inputs[f"member{k}:idx_cb"])
            iglob = self.node_mem2glob[(k, icb)]
            Fnode[iglob, 2] += inputs[f"member{k}:buoyancy_force"]

        # Get transition piece inertial properties
        itrans_platform = util.closest_node(node_platform, inputs["transition_node"])
        m_trans = float(inputs["transition_piece_mass"])
        r_trans = Rnode[itrans_platform]
        I_trans = m_trans * r_trans ** 2.0 * np.r_[0.5, 0.5, 1.0, np.zeros(3)]
        outputs["transition_piece_I"] = I_trans

        # Store outputs
        outputs["platform_Rnode"] = NULL * np.ones(NNODES_MAX)
        outputs["platform_Rnode"][:nnode] = Rnode
        outputs["platform_Fnode"] = NULL * np.ones((NNODES_MAX, 3))
        outputs["platform_Fnode"][:nnode, :] = Fnode

    def set_element_props(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Load in number of members
        opt = self.options["options"]
        n_member = opt["floating"]["members"]["n_members"]

        # Initialize running lists across all members
        elem_D = np.array([])
        elem_t = np.array([])
        elem_A = np.array([])
        elem_Asx = np.array([])
        elem_Asy = np.array([])
        elem_Ixx = np.array([])
        elem_Iyy = np.array([])
        elem_Izz = np.array([])
        elem_rho = np.array([])
        elem_E = np.array([])
        elem_G = np.array([])
        elem_sigy = np.array([])
        elem_Px1 = np.array([])
        elem_Px2 = np.array([])
        elem_Py1 = np.array([])
        elem_Py2 = np.array([])
        elem_Pz1 = np.array([])
        elem_Pz2 = np.array([])
        elem_qdyn = np.array([])
        elem_memid = np.array([], dtype=np.int_)

        mass = 0.0
        m_ball = 0.0
        cost = 0.0
        volume = 0.0
        Awater = 0.0
        Iwater = 0.0
        m_added = np.zeros(6)
        cg_plat = np.zeros(3)
        cb_plat = np.zeros(3)
        centroid = outputs["platform_centroid"][:2]
        variable_capacity = np.zeros(n_member)

        # Append all member data
        for k in range(n_member):
            n = np.where(inputs[f"member{k}:section_A"] == NULL)[0][0]
            elem_D = np.append(elem_D, inputs[f"member{k}:section_D"][:n])
            elem_t = np.append(elem_t, inputs[f"member{k}:section_t"][:n])
            elem_A = np.append(elem_A, inputs[f"member{k}:section_A"][:n])
            elem_Asx = np.append(elem_Asx, inputs[f"member{k}:section_Asx"][:n])
            elem_Asy = np.append(elem_Asy, inputs[f"member{k}:section_Asy"][:n])
            elem_Ixx = np.append(elem_Ixx, inputs[f"member{k}:section_Ixx"][:n])
            elem_Iyy = np.append(elem_Iyy, inputs[f"member{k}:section_Iyy"][:n])
            elem_Izz = np.append(elem_Izz, inputs[f"member{k}:section_Izz"][:n])
            elem_rho = np.append(elem_rho, inputs[f"member{k}:section_rho"][:n])
            elem_E = np.append(elem_E, inputs[f"member{k}:section_E"][:n])
            elem_G = np.append(elem_G, inputs[f"member{k}:section_G"][:n])
            elem_sigy = np.append(elem_sigy, inputs[f"member{k}:section_sigma_y"][:n])
            elem_qdyn = np.append(elem_qdyn, inputs[f"member{k}:qdyn"][:n])
            elem_memid = np.append(elem_memid, k * np.ones(n, dtype=np.int_))

            # The loads should come in with length n+1
            elem_Px1 = np.append(elem_Px1, inputs[f"member{k}:Px"][:n])
            elem_Px2 = np.append(elem_Px2, inputs[f"member{k}:Px"][1 : (n + 1)])
            elem_Py1 = np.append(elem_Py1, inputs[f"member{k}:Py"][:n])
            elem_Py2 = np.append(elem_Py2, inputs[f"member{k}:Py"][1 : (n + 1)])
            elem_Pz1 = np.append(elem_Pz1, inputs[f"member{k}:Pz"][:n])
            elem_Pz2 = np.append(elem_Pz2, inputs[f"member{k}:Pz"][1 : (n + 1)])

            # Mass, volume, cost tallies
            imass = inputs[f"member{k}:total_mass"]
            ivol = inputs[f"member{k}:displacement"]

            mass += imass
            volume += ivol
            cost += inputs[f"member{k}:total_cost"]
            m_ball += inputs[f"member{k}:ballast_mass"]
            Awater_k = inputs[f"member{k}:Awater"]
            Awater += Awater_k
            Rwater2 = np.sum((inputs[f"member{k}:waterline_centroid"] - centroid) ** 2)
            Iwater += inputs[f"member{k}:Iwater"] + Awater_k * Rwater2
            m_added += inputs[f"member{k}:added_mass"]
            variable_capacity[k] = inputs[f"member{k}:variable_ballast_capacity"]

            # Center of mass / buoyancy tallies
            cg_plat += imass * inputs[f"member{k}:center_of_mass"]
            cb_plat += ivol * inputs[f"member{k}:center_of_buoyancy"]

        # Add transition piece
        m_trans = inputs["transition_piece_mass"]
        cg_trans = inputs["transition_node"]
        I_trans = util.assembleI(outputs["transition_piece_I"])
        mass += m_trans
        cost += inputs["transition_piece_cost"]
        cg_plat += m_trans * cg_trans

        # Finalize outputs
        cg_plat /= mass
        cb_plat /= volume

        # With CG known, loop back through to compute platform I
        unit_z = np.array([0.0, 0.0, 1.0])
        I_hull = np.zeros((3, 3))
        for k in range(n_member):
            xyz_k = inputs[f"member{k}:nodes_xyz"]
            inodes = np.where(xyz_k[:, 0] == NULL)[0][0]
            xyz_k = xyz_k[:inodes, :]

            imass = inputs[f"member{k}:total_mass"]
            cg_k = inputs[f"member{k}:center_of_mass"]
            R = cg_plat - cg_k

            # Figure out angle to make member parallel to global c.s.
            vec_k = xyz_k[-1, :] - xyz_k[0, :]
            T = util.rotate_align_vectors(vec_k, unit_z)

            # Rotate member inertia tensor
            I_k = util.assembleI(inputs[f"member{k}:I_total"])
            I_k_rot = T @ I_k @ T.T

            # Now do parallel axis theorem
            I_hull += np.array(I_k_rot) + imass * (np.dot(R, R) * np.eye(3) - np.outer(R, R))

        # Add in transition piece
        R = cg_plat - cg_trans
        I_hull += I_trans + m_trans * (np.dot(R, R) * np.eye(3) - np.outer(R, R))

        # Store outputs
        nelem = elem_A.size
        outputs["platform_elem_D"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_t"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_A"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_Asx"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_Asy"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_Ixx"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_Iyy"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_Izz"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_rho"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_E"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_G"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_sigma_y"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_Px1"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_Px2"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_Py1"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_Py2"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_Pz1"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_Pz2"] = NULL * np.ones(NELEM_MAX)
        outputs["platform_elem_qdyn"] = NULL * np.ones(NELEM_MAX)

        outputs["platform_elem_D"][:nelem] = elem_D
        outputs["platform_elem_t"][:nelem] = elem_t
        outputs["platform_elem_A"][:nelem] = elem_A
        outputs["platform_elem_Asx"][:nelem] = elem_Asx
        outputs["platform_elem_Asy"][:nelem] = elem_Asy
        outputs["platform_elem_Ixx"][:nelem] = elem_Ixx
        outputs["platform_elem_Iyy"][:nelem] = elem_Iyy
        outputs["platform_elem_Izz"][:nelem] = elem_Izz
        outputs["platform_elem_rho"][:nelem] = elem_rho
        outputs["platform_elem_E"][:nelem] = elem_E
        outputs["platform_elem_G"][:nelem] = elem_G
        outputs["platform_elem_sigma_y"][:nelem] = elem_sigy
        outputs["platform_elem_Px1"][:nelem] = elem_Px1
        outputs["platform_elem_Px2"][:nelem] = elem_Px2
        outputs["platform_elem_Py1"][:nelem] = elem_Py1
        outputs["platform_elem_Py2"][:nelem] = elem_Py2
        outputs["platform_elem_Pz1"][:nelem] = elem_Pz1
        outputs["platform_elem_Pz2"][:nelem] = elem_Pz2
        outputs["platform_elem_qdyn"][:nelem] = elem_qdyn
        discrete_outputs["platform_elem_memid"] = elem_memid

        outputs["platform_mass"] = mass
        outputs["platform_ballast_mass"] = m_ball
        outputs["platform_hull_mass"] = mass - m_ball
        outputs["platform_cost"] = cost
        outputs["platform_displacement"] = volume
        outputs["platform_hull_center_of_mass"] = cg_plat
        outputs["platform_center_of_buoyancy"] = cb_plat
        outputs["platform_I_hull"] = util.unassembleI(I_hull)
        outputs["platform_Awater"] = Awater
        outputs["platform_Iwater"] = Iwater
        outputs["platform_added_mass"] = m_added
        outputs["platform_variable_capacity"] = variable_capacity


class TowerPreMember(om.ExplicitComponent):
    def setup(self):
        self.add_input("transition_node", np.zeros(3), units="m")
        self.add_input("tower_height", 0.0, units="m")
        self.add_output("tower_top_node", np.zeros(3), units="m")

    def compute(self, inputs, outputs):
        transition_node = inputs["transition_node"]
        tower_top_node = 0  # previous code altered the original definition of transition_node
        tower_top_node += transition_node
        tower_top_node[2] += float(inputs["tower_height"])
        outputs["tower_top_node"] = tower_top_node


class PlatformTowerFrame(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("options")

    def setup(self):
        opt = self.options["options"]
        n_member = opt["floating"]["members"]["n_members"]
        n_attach = opt["mooring"]["n_attach"]

        self.add_input("platform_nodes", NULL * np.ones((NNODES_MAX, 3)), units="m")
        self.add_input("platform_Fnode", NULL * np.ones((NNODES_MAX, 3)), units="N")
        self.add_input("platform_Rnode", NULL * np.ones(NNODES_MAX), units="m")
        self.add_input("platform_elem_n1", NULL * np.ones(NELEM_MAX, dtype=np.int_))
        self.add_input("platform_elem_n2", NULL * np.ones(NELEM_MAX, dtype=np.int_))
        self.add_input("platform_elem_D", NULL * np.ones(NELEM_MAX), units="m")
        self.add_input("platform_elem_t", NULL * np.ones(NELEM_MAX), units="m")
        self.add_input("platform_elem_A", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_input("platform_elem_Asx", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_input("platform_elem_Asy", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_input("platform_elem_Ixx", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_input("platform_elem_Iyy", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_input("platform_elem_Izz", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_input("platform_elem_rho", NULL * np.ones(NELEM_MAX), units="kg/m**3")
        self.add_input("platform_elem_E", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_input("platform_elem_G", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_input("platform_elem_sigma_y", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_input("platform_elem_Px1", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_input("platform_elem_Px2", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_input("platform_elem_Py1", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_input("platform_elem_Py2", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_input("platform_elem_Pz1", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_input("platform_elem_Pz2", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_input("platform_elem_qdyn", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_input("platform_hull_center_of_mass", np.zeros(3), units="m")
        self.add_input("platform_mass", 0.0, units="kg")
        self.add_input("platform_I_hull", np.zeros(6), units="kg*m**2")
        self.add_input("platform_displacement", 0.0, units="m**3")

        self.add_input("tower_nodes", NULL * np.ones((MEMMAX, 3)), units="m")
        self.add_output("tower_Fnode", copy_shape="tower_nodes", units="N")
        self.add_input("tower_Rnode", NULL * np.ones(MEMMAX), units="m")
        self.add_output("tower_elem_n1", copy_shape="tower_elem_A")
        self.add_output("tower_elem_n2", copy_shape="tower_elem_A")
        self.add_output("tower_elem_L", copy_shape="tower_elem_A", units="m")
        self.add_input("tower_elem_D", NULL * np.ones(MEMMAX), units="m")
        self.add_input("tower_elem_t", NULL * np.ones(MEMMAX), units="m")
        self.add_input("tower_elem_A", NULL * np.ones(MEMMAX), units="m**2")
        self.add_input("tower_elem_Asx", NULL * np.ones(MEMMAX), units="m**2")
        self.add_input("tower_elem_Asy", NULL * np.ones(MEMMAX), units="m**2")
        self.add_input("tower_elem_Ixx", NULL * np.ones(MEMMAX), units="kg*m**2")
        self.add_input("tower_elem_Iyy", NULL * np.ones(MEMMAX), units="kg*m**2")
        self.add_input("tower_elem_Izz", NULL * np.ones(MEMMAX), units="kg*m**2")
        self.add_input("tower_elem_rho", NULL * np.ones(MEMMAX), units="kg/m**3")
        self.add_input("tower_elem_E", NULL * np.ones(MEMMAX), units="Pa")
        self.add_input("tower_elem_G", NULL * np.ones(MEMMAX), units="Pa")
        self.add_input("tower_elem_sigma_y", NULL * np.ones(MEMMAX), units="Pa")
        self.add_input("tower_elem_Px", NULL * np.ones(MEMMAX), units="N/m")
        self.add_output("tower_elem_Px1", NULL * np.ones(MEMMAX), units="N/m")
        self.add_output("tower_elem_Px2", NULL * np.ones(MEMMAX), units="N/m")
        self.add_input("tower_elem_Py", NULL * np.ones(MEMMAX), units="N/m")
        self.add_output("tower_elem_Py1", NULL * np.ones(MEMMAX), units="N/m")
        self.add_output("tower_elem_Py2", NULL * np.ones(MEMMAX), units="N/m")
        self.add_input("tower_elem_Pz", NULL * np.ones(MEMMAX), units="N/m")
        self.add_output("tower_elem_Pz1", NULL * np.ones(MEMMAX), units="N/m")
        self.add_output("tower_elem_Pz2", NULL * np.ones(MEMMAX), units="N/m")
        self.add_input("tower_elem_qdyn", NULL * np.ones(MEMMAX), units="Pa")
        self.add_input("tower_center_of_mass", np.zeros(3), units="m")
        self.add_input("tower_mass", 0.0, units="kg")

        self.add_input("rho_water", 0.0, units="kg/m**3")
        self.add_input("tower_top_node", np.zeros(3), units="m")
        self.add_input("transition_node", np.zeros(3), units="m")
        self.add_input("rna_mass", 0.0, units="kg")
        self.add_input("rna_cg", np.zeros(3), units="m")
        self.add_input("mooring_neutral_load", np.zeros((n_attach, 3)), units="N")
        self.add_input("platform_variable_capacity", np.zeros(n_member), units="m**3")

        for k in range(n_member):
            self.add_input(f"member{k}:nodes_xyz", NULL * np.ones((MEMMAX, 3)), units="m")
            self.add_input(f"member{k}:variable_ballast_Vpts", val=np.zeros(10), units="m**3")
            self.add_input(f"member{k}:variable_ballast_spts", val=np.zeros(10))

        self.add_output("system_nodes", NULL * np.ones((NNODES_MAX, 3)), units="m")
        self.add_output("system_Fnode", NULL * np.ones((NNODES_MAX, 3)), units="N")
        self.add_output("system_Rnode", NULL * np.ones(NNODES_MAX), units="m")
        self.add_output("system_elem_n1", NULL * np.ones(NELEM_MAX, dtype=np.int_))
        self.add_output("system_elem_n2", NULL * np.ones(NELEM_MAX, dtype=np.int_))
        self.add_output("system_elem_L", NULL * np.ones(NELEM_MAX), units="m")
        self.add_output("system_elem_D", NULL * np.ones(NELEM_MAX), units="m")
        self.add_output("system_elem_t", NULL * np.ones(NELEM_MAX), units="m")
        self.add_output("system_elem_A", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_output("system_elem_Asx", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_output("system_elem_Asy", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_output("system_elem_Ixx", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_output("system_elem_Iyy", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_output("system_elem_Izz", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_output("system_elem_rho", NULL * np.ones(NELEM_MAX), units="kg/m**3")
        self.add_output("system_elem_E", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_output("system_elem_G", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_output("system_elem_sigma_y", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_output("system_elem_Px1", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_output("system_elem_Px2", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_output("system_elem_Py1", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_output("system_elem_Py2", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_output("system_elem_Pz1", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_output("system_elem_Pz2", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_output("system_elem_qdyn", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_output("system_structural_center_of_mass", np.zeros(3), units="m")
        self.add_output("system_structural_mass", 0.0, units="kg")
        self.add_output("system_center_of_mass", np.zeros(3), units="m")
        self.add_output("system_mass", 0.0, units="kg")
        self.add_output("variable_ballast_mass", 0.0, units="kg")
        self.add_output("variable_center_of_mass", val=np.zeros(3), units="m")
        self.add_output("constr_variable_margin", val=0.0)
        self.add_output("member_variable_volume", val=np.zeros(n_member), units="m**3")
        self.add_output("member_variable_height", val=np.zeros(n_member))
        self.add_output("platform_total_center_of_mass", np.zeros(3), units="m")
        self.add_output("platform_I_total", np.zeros(6), units="kg*m**2")

    def compute(self, inputs, outputs):
        # Combine nodes
        node_platform = inputs["platform_nodes"]
        node_tower = inputs["tower_nodes"]

        nnode_platform = np.where(node_platform[:, 0] == NULL)[0][0]
        nnode_tower = np.where(node_tower[:, 0] == NULL)[0][0]
        nnode_system = nnode_platform + np.maximum(1, nnode_tower) - 1

        nelem_platform = np.where(inputs["platform_elem_A"] == NULL)[0][0]
        nelem_tower = np.where(inputs["tower_elem_A"] == NULL)[0][0]
        nelem_system = nelem_platform + nelem_tower

        # Combine elements indices and have tower base node point to platform transition node
        outputs["tower_Fnode"] = np.zeros(node_tower.shape)
        outputs["tower_elem_n1"] = NULL * np.ones(MEMMAX, dtype=np.int_)
        outputs["tower_elem_n2"] = NULL * np.ones(MEMMAX, dtype=np.int_)
        outputs["tower_elem_L"] = NULL * np.ones(MEMMAX)
        tower_n1 = np.arange(nelem_tower, dtype=np.int_)
        tower_n2 = np.arange(nelem_tower, dtype=np.int_) + 1
        outputs["tower_elem_n1"][:nelem_tower] = idx1 = tower_n1.copy()
        outputs["tower_elem_n2"][:nelem_tower] = idx2 = tower_n2.copy()
        itrans_platform = util.closest_node(node_platform[:nnode_platform, :], inputs["transition_node"])
        tower_n1 += nnode_platform - 1
        tower_n2 += nnode_platform - 1
        tower_n1[0] = itrans_platform

        outputs["tower_elem_L"][:nelem_tower] = np.sqrt(
            np.sum((node_tower[idx2, :] - node_tower[idx1, :]) ** 2, axis=1)
        )

        # Store all outputs
        outputs["system_nodes"] = NULL * np.ones((NNODES_MAX, 3))
        outputs["system_Fnode"] = NULL * np.ones((NNODES_MAX, 3))
        outputs["system_Rnode"] = NULL * np.ones(NNODES_MAX)
        outputs["system_elem_n1"] = NULL * np.ones(NELEM_MAX, dtype=np.int_)
        outputs["system_elem_n2"] = NULL * np.ones(NELEM_MAX, dtype=np.int_)
        outputs["system_elem_L"] = NULL * np.ones(NELEM_MAX)

        outputs["system_nodes"][:nnode_system, :] = sysnode = np.vstack(
            (node_platform[:nnode_platform, :], node_tower[1:nnode_tower, :])
        )
        outputs["system_Fnode"][:nnode_system, :] = np.vstack(
            (inputs["platform_Fnode"][:nnode_platform, :], outputs["tower_Fnode"][1:nnode_tower, :])
        )
        outputs["system_Rnode"][:nnode_system] = np.r_[
            inputs["platform_Rnode"][:nnode_platform], inputs["tower_Rnode"][1:nnode_tower]
        ]

        outputs["system_elem_n1"][:nelem_system] = idx1 = np.r_[
            inputs["platform_elem_n1"][:nelem_platform],
            tower_n1,
        ]
        outputs["system_elem_n2"][:nelem_system] = idx2 = np.r_[
            inputs["platform_elem_n2"][:nelem_platform],
            tower_n2,
        ]

        outputs["system_elem_L"][:nelem_system] = np.sqrt(
            np.sum((sysnode[np.int_(idx2), :] - sysnode[np.int_(idx1), :]) ** 2, axis=1)
        )

        for var in [
            "elem_D",
            "elem_t",
            "elem_A",
            "elem_Asx",
            "elem_Asy",
            "elem_Ixx",
            "elem_Iyy",
            "elem_Izz",
            "elem_rho",
            "elem_E",
            "elem_G",
            "elem_sigma_y",
            "elem_qdyn",
        ]:
            outputs["system_" + var] = NULL * np.ones(NELEM_MAX)
            outputs["system_" + var][:nelem_system] = np.r_[
                inputs["platform_" + var][:nelem_platform], inputs["tower_" + var][:nelem_tower]
            ]

        # Have to divide up tower member loads to beginning and end points
        for var in ["elem_Px1", "elem_Py1", "elem_Pz1", "elem_Px2", "elem_Py2", "elem_Pz2"]:
            outputs["system_" + var] = NULL * np.ones(NELEM_MAX)
            outputs["tower_" + var] = NULL * np.ones(MEMMAX)
            tower_P = inputs["tower_" + var[:-1]]
            outputs["tower_" + var][:nelem_tower] = (
                tower_P[:nelem_tower] if var[-1] == "1" else tower_P[1 : (nelem_tower + 1)]
            )
            outputs["system_" + var][:nelem_system] = np.r_[
                inputs["platform_" + var][:nelem_platform], outputs["tower_" + var][:nelem_tower]
            ]

        # Mass summaries
        m_platform = inputs["platform_mass"]
        cg_platform = inputs["platform_hull_center_of_mass"]
        I_platform = util.assembleI(inputs["platform_I_hull"])
        m_tower = inputs["tower_mass"]
        m_rna = inputs["rna_mass"]
        m_sys = m_platform + m_tower + m_rna
        outputs["system_structural_mass"] = m_sys

        outputs["system_structural_center_of_mass"] = (
            m_platform * cg_platform
            + m_tower * inputs["tower_center_of_mass"]
            + m_rna * (inputs["rna_cg"] + inputs["tower_top_node"])
        ) / m_sys

        # Balance out variable ballast
        mooringFz = inputs["mooring_neutral_load"][:, 2].sum()
        capacity = inputs["platform_variable_capacity"]
        capacity_sum = capacity.sum() + EPS  # Avoid divide by zeros
        rho_water = inputs["rho_water"]
        m_variable = inputs["platform_displacement"] * rho_water - m_sys + mooringFz / gravity
        V_variable = m_variable / rho_water
        outputs["variable_ballast_mass"] = m_variable
        outputs["constr_variable_margin"] = V_variable / capacity_sum
        V_variable_member = V_variable * capacity / capacity_sum
        outputs["member_variable_volume"] = V_variable_member
        m_variable_member = V_variable_member * rho_water

        # Now find the CG of the variable mass assigned to each member
        n_member = capacity.size
        outputs["member_variable_height"] = np.zeros(n_member)
        cg_variable_member = np.zeros((n_member, 3))
        for k in range(n_member):
            if V_variable_member[k] == 0.0:
                continue

            xyz = inputs[f"member{k}:nodes_xyz"]
            inodes = np.where(xyz[:, 0] == NULL)[0][0]
            xyz = xyz[:inodes, :]
            dxyz = xyz[-1, :] - xyz[0, :]

            spts = inputs[f"member{k}:variable_ballast_spts"]
            Vpts = inputs[f"member{k}:variable_ballast_Vpts"]

            s_cg = np.interp(0.5 * V_variable_member[k], Vpts, spts)
            cg_variable_member[k, :] = xyz[0, :] + s_cg * dxyz

            s_end = np.interp(V_variable_member[k], Vpts, spts)
            outputs["member_variable_height"][k] = s_end - spts[0]

        cg_variable = np.dot(V_variable_member, cg_variable_member) / V_variable
        outputs["variable_center_of_mass"] = cg_variable

        # Now find total system mass
        outputs["system_mass"] = m_sys + m_variable
        outputs["system_center_of_mass"] = (
            m_sys * outputs["system_structural_center_of_mass"] + m_variable * cg_variable
        ) / (m_sys + m_variable)

        # Compute the total cg for the platform and the variable ballast together using a weighted sum approach
        cg_plat_total = (m_variable * cg_variable + m_platform * cg_platform) / (m_variable + m_platform)
        outputs["platform_total_center_of_mass"] = cg_plat_total

        # Now loop again to compute variable I
        unit_z = np.array([0.0, 0.0, 1.0])
        I_variable = np.zeros((3, 3))
        for k in range(n_member):
            if V_variable_member[k] == 0.0:
                continue

            xyz = inputs[f"member{k}:nodes_xyz"]
            inodes = np.where(xyz[:, 0] == NULL)[0][0]
            xyz = xyz[:inodes, :]
            vec_k = xyz[-1, :] - xyz[0, :]

            ds = outputs["member_variable_height"][k]

            # Compute I aligned with member
            h_k = ds * np.sqrt(np.sum(vec_k ** 2))
            if h_k == 0.0:
                continue
            r_k = np.sqrt(V_variable_member[k] / h_k / np.pi)
            I_k = (
                m_variable_member[k] * np.r_[(3 * r_k ** 2 + h_k ** 2) / 12.0 * np.ones(2), 0.5 * r_k ** 2, np.ones(3)]
            )

            # Rotate I to global c.s.
            T = util.rotate_align_vectors(vec_k, unit_z)
            I_k_rot = T @ util.assembleI(I_k) @ T.T

            # Now do parallel axis theorem
            R = cg_variable - cg_variable_member[k, :]
            I_variable += np.array(I_k_rot) + m_variable_member[k] * (np.dot(R, R) * np.eye(3) - np.outer(R, R))

        # Find platform I with variable contribution
        I_total = np.zeros((3, 3))

        # Compute the full moment of inertia for the platform and variable ballast
        R = cg_plat_total - cg_platform
        I_total += I_platform + m_platform * (np.dot(R, R) * np.eye(3) - np.outer(R, R))

        R = cg_plat_total - cg_variable
        I_total += I_variable + m_variable * (np.dot(R, R) * np.eye(3) - np.outer(R, R))
        outputs["platform_I_total"] = util.unassembleI(I_total)


class FrameAnalysis(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("options")

    def setup(self):
        opt = self.options["options"]
        n_attach = opt["mooring"]["n_attach"]

        self.add_input("platform_mass", 0.0, units="kg")
        self.add_input("platform_hull_center_of_mass", np.zeros(3), units="m")
        self.add_input("platform_added_mass", np.zeros(6), units="kg")
        self.add_input("platform_center_of_buoyancy", np.zeros(3), units="m")
        self.add_input("platform_displacement", 0.0, units="m**3")

        self.add_input("tower_nodes", NULL * np.ones((MEMMAX, 3)), units="m")
        self.add_input("tower_Fnode", NULL * np.ones((MEMMAX, 3)), units="N")
        self.add_input("tower_Rnode", NULL * np.ones(MEMMAX), units="m")
        self.add_input("tower_elem_n1", NULL * np.ones(MEMMAX))
        self.add_input("tower_elem_n2", NULL * np.ones(MEMMAX))
        self.add_input("tower_elem_D", NULL * np.ones(MEMMAX), units="m")
        self.add_input("tower_elem_t", NULL * np.ones(MEMMAX), units="m")
        self.add_input("tower_elem_L", NULL * np.ones(MEMMAX), units="m")
        self.add_input("tower_elem_A", NULL * np.ones(MEMMAX), units="m**2")
        self.add_input("tower_elem_Asx", NULL * np.ones(MEMMAX), units="m**2")
        self.add_input("tower_elem_Asy", NULL * np.ones(MEMMAX), units="m**2")
        self.add_input("tower_elem_Ixx", NULL * np.ones(MEMMAX), units="kg*m**2")
        self.add_input("tower_elem_Iyy", NULL * np.ones(MEMMAX), units="kg*m**2")
        self.add_input("tower_elem_Izz", NULL * np.ones(MEMMAX), units="kg*m**2")
        self.add_input("tower_elem_rho", NULL * np.ones(MEMMAX), units="kg/m**3")
        self.add_input("tower_elem_E", NULL * np.ones(MEMMAX), units="Pa")
        self.add_input("tower_elem_G", NULL * np.ones(MEMMAX), units="Pa")
        self.add_input("tower_elem_Px1", NULL * np.ones(MEMMAX), units="N/m")
        self.add_input("tower_elem_Px2", NULL * np.ones(MEMMAX), units="N/m")
        self.add_input("tower_elem_Py1", NULL * np.ones(MEMMAX), units="N/m")
        self.add_input("tower_elem_Py2", NULL * np.ones(MEMMAX), units="N/m")
        self.add_input("tower_elem_Pz1", NULL * np.ones(MEMMAX), units="N/m")
        self.add_input("tower_elem_Pz2", NULL * np.ones(MEMMAX), units="N/m")

        self.add_input("system_nodes", NULL * np.ones((NNODES_MAX, 3)), units="m")
        self.add_input("system_Fnode", NULL * np.ones((NNODES_MAX, 3)), units="N")
        self.add_input("system_Rnode", NULL * np.ones(NNODES_MAX), units="m")
        self.add_input("system_elem_n1", NULL * np.ones(NELEM_MAX, dtype=np.int_))
        self.add_input("system_elem_n2", NULL * np.ones(NELEM_MAX, dtype=np.int_))
        self.add_input("system_elem_D", NULL * np.ones(NELEM_MAX), units="m")
        self.add_input("system_elem_t", NULL * np.ones(NELEM_MAX), units="m")
        self.add_input("system_elem_L", NULL * np.ones(NELEM_MAX), units="m")
        self.add_input("system_elem_A", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_input("system_elem_Asx", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_input("system_elem_Asy", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_input("system_elem_Ixx", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_input("system_elem_Iyy", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_input("system_elem_Izz", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_input("system_elem_rho", NULL * np.ones(NELEM_MAX), units="kg/m**3")
        self.add_input("system_elem_E", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_input("system_elem_G", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_input("system_elem_Px1", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_input("system_elem_Px2", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_input("system_elem_Py1", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_input("system_elem_Py2", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_input("system_elem_Pz1", NULL * np.ones(NELEM_MAX), units="N/m")
        self.add_input("system_elem_Pz2", NULL * np.ones(NELEM_MAX), units="N/m")

        self.add_input("transition_node", np.zeros(3), units="m")
        self.add_input("transition_piece_mass", 0.0, units="kg")
        self.add_input("transition_piece_I", np.zeros(6), units="kg*m**2")
        self.add_input("rna_F", np.zeros(3), units="N")
        self.add_input("rna_M", np.zeros(3), units="N*m")
        self.add_input("mooring_neutral_load", np.zeros((n_attach, 3)), units="N")
        self.add_input("mooring_fairlead_joints", np.zeros((n_attach, 3)), units="m")
        self.add_input("mooring_stiffness", np.zeros((6, 6)), units="N/m")
        self.add_input("variable_ballast_mass", 0.0, units="kg")
        self.add_input("variable_center_of_mass", val=np.zeros(3), units="m")

        NFREQ2 = int(NFREQ / 2)
        self.add_output("tower_freqs", val=np.zeros(NFREQ), units="Hz")
        self.add_output("tower_fore_aft_modes", val=np.zeros((NFREQ2, 5)))
        self.add_output("tower_side_side_modes", val=np.zeros((NFREQ2, 5)))
        self.add_output("tower_torsion_modes", val=np.zeros((NFREQ2, 5)))
        self.add_output("tower_fore_aft_freqs", val=np.zeros(NFREQ2))
        self.add_output("tower_side_side_freqs", val=np.zeros(NFREQ2))
        self.add_output("tower_torsion_freqs", val=np.zeros(NFREQ2))

        self.add_output("system_base_F", np.zeros(3), units="N")
        self.add_output("system_base_M", np.zeros(3), units="N*m")
        self.add_output("system_Fz", NULL * np.ones(NELEM_MAX), units="N")
        self.add_output("system_Vx", NULL * np.ones(NELEM_MAX), units="N")
        self.add_output("system_Vy", NULL * np.ones(NELEM_MAX), units="N")
        self.add_output("system_Mxx", NULL * np.ones(NELEM_MAX), units="N*m")
        self.add_output("system_Myy", NULL * np.ones(NELEM_MAX), units="N*m")
        self.add_output("system_Mzz", NULL * np.ones(NELEM_MAX), units="N*m")

        self.add_output("tower_base_F", np.zeros(3), units="N")
        self.add_output("tower_base_M", np.zeros(3), units="N*m")
        self.add_output("tower_Fz", NULL * np.ones(NELEM_MAX), units="N")
        self.add_output("tower_Vx", NULL * np.ones(NELEM_MAX), units="N")
        self.add_output("tower_Vy", NULL * np.ones(NELEM_MAX), units="N")
        self.add_output("tower_Mxx", NULL * np.ones(NELEM_MAX), units="N*m")
        self.add_output("tower_Myy", NULL * np.ones(NELEM_MAX), units="N*m")
        self.add_output("tower_Mzz", NULL * np.ones(NELEM_MAX), units="N*m")

    def compute(self, inputs, outputs):

        # Unpack variables
        opt = self.options["options"]
        n_attach = opt["mooring"]["n_attach"]
        cb = inputs["platform_center_of_buoyancy"]
        V_total = inputs["platform_displacement"]
        I_trans = inputs["transition_piece_I"]
        m_variable = float(inputs["variable_ballast_mass"])
        cg_variable = inputs["variable_center_of_mass"]

        fairlead_joints = inputs["mooring_fairlead_joints"]
        mooringF = inputs["mooring_neutral_load"]
        mooringK = np.abs(np.diag(inputs["mooring_stiffness"]))

        # Create frame3dd instance: nodes, elements, reactions, and options
        for frame in ["tower", "system"]:
            nodes = inputs[f"{frame}_nodes"]
            nnode = np.where(nodes[:, 0] == NULL)[0][0]
            nodes = nodes[:nnode, :]
            rnode = np.zeros(nnode)  # inputs[f"{frame}_Rnode"][:nnode]
            Fnode = inputs[f"{frame}_Fnode"][:nnode, :]
            Mnode = np.zeros((nnode, 3))
            ihub = np.argmax(nodes[:, 2]) - 1
            itrans = util.closest_node(nodes, inputs["transition_node"])

            N1 = np.int_(inputs[f"{frame}_elem_n1"])
            nelem = np.where(N1 == NULL)[0][0]
            N1 = N1[:nelem]
            N2 = np.int_(inputs[f"{frame}_elem_n2"][:nelem])
            A = inputs[f"{frame}_elem_A"][:nelem]
            Asx = inputs[f"{frame}_elem_Asx"][:nelem]
            Asy = inputs[f"{frame}_elem_Asy"][:nelem]
            Ixx = inputs[f"{frame}_elem_Ixx"][:nelem]
            Iyy = inputs[f"{frame}_elem_Iyy"][:nelem]
            Izz = inputs[f"{frame}_elem_Izz"][:nelem]
            rho = inputs[f"{frame}_elem_rho"][:nelem]
            E = inputs[f"{frame}_elem_E"][:nelem]
            G = inputs[f"{frame}_elem_G"][:nelem]
            roll = np.zeros(nelem)
            L = inputs[f"{frame}_elem_L"][:nelem]  # np.sqrt(np.sum((nodes[N2, :] - nodes[N1, :]) ** 2, axis=1))

            inodes = np.arange(nnode) + 1
            node_obj = pyframe3dd.NodeData(inodes, nodes[:, 0], nodes[:, 1], nodes[:, 2], rnode)

            ielem = np.arange(nelem) + 1
            elem_obj = pyframe3dd.ElementData(ielem, N1 + 1, N2 + 1, A, Asx, Asy, Izz, Ixx, Iyy, E, G, roll, rho)

            # Use Mooring stiffness (TODO Hydro_K too)
            if frame == "tower":
                rid = np.array([itrans])  # np.array([np.argmin(nodes[:, 2])])
            else:
                ind = []
                for k in range(n_attach):
                    ind.append(util.closest_node(nodes, fairlead_joints[k, :]))
                rid = np.array([ind])  # np.array([np.argmin(nodes[:, 2])])

            Rx = Ry = Rz = Rxx = Ryy = Rzz = RIGID * np.ones(rid.size)
            # Rx, Ry, Rz = [mooringK[0]], [mooringK[1]], [mooringK[2]]
            # Only this solution works and there isn't much different with fully rigid
            # Rx, Ry, Rz = [RIGID], [RIGID], [mooringK[2]]
            # Rxx, Ryy, Rzz = [RIGID], [RIGID], [RIGID]
            react_obj = pyframe3dd.ReactionData(rid + 1, Rx, Ry, Rz, Rxx, Ryy, Rzz, rigid=RIGID)

            frame3dd_opt = opt["WISDEM"]["FloatingSE"]["frame3dd"]
            opt_obj = pyframe3dd.Options(frame3dd_opt["shear"], frame3dd_opt["geom"], -1.0)

            myframe = pyframe3dd.Frame(node_obj, react_obj, elem_obj, opt_obj)

            # Added mass
            m_trans = float(inputs["transition_piece_mass"])
            if frame == "tower":
                m_trans += float(inputs["platform_mass"]) + inputs["platform_added_mass"][0] + m_variable
                cg_trans = inputs["transition_node"] - inputs["platform_hull_center_of_mass"]
                I_trans[:3] += inputs["platform_added_mass"][3:]

            else:
                m_trans += m_variable
                cg_trans = np.zeros(3)
            add_gravity = True
            mID = np.array([itrans], dtype=np.int_).flatten()
            m_add = np.array([m_trans]).flatten()
            I_add = np.c_[I_trans]
            cg_add = np.c_[cg_trans]
            myframe.changeExtraNodeMass(
                mID + 1,
                m_add,
                I_add[0, :],
                I_add[1, :],
                I_add[2, :],
                I_add[3, :],
                I_add[4, :],
                I_add[5, :],
                cg_add[0, :],
                cg_add[1, :],
                cg_add[2, :],
                add_gravity,
            )

            # Dynamics
            if frame == "tower" and frame3dd_opt["modal"]:
                Mmethod = 1
                lump = 0
                shift = 0.0
                myframe.enableDynamics(2 * NFREQ, Mmethod, lump, frame3dd_opt["tol"], shift)

            # Initialize loading with gravity, mooring line forces, and buoyancy (already in nodal forces)
            gx = gy = 0.0
            gz = -gravity
            load_obj = pyframe3dd.StaticLoadCase(gx, gy, gz)

            if frame == "system":
                for k in range(n_attach):
                    ind = util.closest_node(nodes, fairlead_joints[k, :])
                    Fnode[ind, :] += mooringF[k, :]
            else:
                # Combine all buoyancy forces into one
                ind = util.closest_node(nodes, cb)
                Fnode[ind, -1] += V_total * 1025 * gravity

            Fnode[ihub, :] += inputs["rna_F"]
            Mnode[ihub, :] += inputs["rna_M"]
            nF = np.where(np.abs(Fnode).sum(axis=1) > 0.0)[0]
            load_obj.changePointLoads(
                nF + 1, Fnode[nF, 0], Fnode[nF, 1], Fnode[nF, 2], Mnode[nF, 0], Mnode[nF, 1], Mnode[nF, 2]
            )

            # trapezoidally distributed loads
            xx1 = xy1 = xz1 = np.zeros(ielem.size)
            xx2 = xy2 = xz2 = 0.99 * L  # multiply slightly less than unity b.c. of precision
            wx1 = inputs[f"{frame}_elem_Px1"][:nelem]
            wx2 = inputs[f"{frame}_elem_Px2"][:nelem]
            wy1 = inputs[f"{frame}_elem_Py1"][:nelem]
            wy2 = inputs[f"{frame}_elem_Py2"][:nelem]
            wz1 = inputs[f"{frame}_elem_Pz1"][:nelem]
            wz2 = inputs[f"{frame}_elem_Pz2"][:nelem]
            load_obj.changeTrapezoidalLoads(ielem, xx1, xx2, wx1, wx2, xy1, xy2, wy1, wy2, xz1, xz2, wz1, wz2)

            # Add the load case and run
            myframe.addLoadCase(load_obj)
            # myframe.write(f"{frame}.3dd")
            # myframe.draw()
            displacements, forces, reactions, internalForces, mass, modal = myframe.run()

            # natural frequncies
            if frame == "tower" and frame3dd_opt["modal"]:
                outputs[f"{frame}_freqs"] = modal.freq[:NFREQ]

                # Get all mode shapes in batch
                NFREQ2 = int(NFREQ / 2)
                freq_x, freq_y, freq_z, mshapes_x, mshapes_y, mshapes_z = util.get_xyz_mode_shapes(
                    nodes[:, 2], modal.freq, modal.xdsp, modal.ydsp, modal.zdsp, modal.xmpf, modal.ympf, modal.zmpf
                )
                outputs[f"{frame}_fore_aft_freqs"] = freq_x[:NFREQ2]
                outputs[f"{frame}_side_side_freqs"] = freq_y[:NFREQ2]
                outputs[f"{frame}_torsion_freqs"] = freq_z[:NFREQ2]
                outputs[f"{frame}_fore_aft_modes"] = mshapes_x[:NFREQ2, :]
                outputs[f"{frame}_side_side_modes"] = mshapes_y[:NFREQ2, :]
                outputs[f"{frame}_torsion_modes"] = mshapes_z[:NFREQ2, :]

            # Determine reaction forces
            outputs[f"{frame}_base_F"] = -np.array([reactions.Fx.sum(), reactions.Fy.sum(), reactions.Fz.sum()])
            outputs[f"{frame}_base_M"] = -np.array([reactions.Mxx.sum(), reactions.Myy.sum(), reactions.Mzz.sum()])

            # Forces and moments along the structure
            ic = 0  # case number
            outputs[f"{frame}_Fz"] = NULL * np.ones(NELEM_MAX)
            outputs[f"{frame}_Vx"] = NULL * np.ones(NELEM_MAX)
            outputs[f"{frame}_Vy"] = NULL * np.ones(NELEM_MAX)
            outputs[f"{frame}_Mxx"] = NULL * np.ones(NELEM_MAX)
            outputs[f"{frame}_Myy"] = NULL * np.ones(NELEM_MAX)
            outputs[f"{frame}_Mzz"] = NULL * np.ones(NELEM_MAX)

            outputs[f"{frame}_Fz"][:nelem] = forces.Nx[ic, 1::2]
            outputs[f"{frame}_Vx"][:nelem] = -forces.Vz[ic, 1::2]
            outputs[f"{frame}_Vy"][:nelem] = forces.Vy[ic, 1::2]
            outputs[f"{frame}_Mxx"][:nelem] = -forces.Mzz[ic, 1::2]
            outputs[f"{frame}_Myy"][:nelem] = forces.Myy[ic, 1::2]
            outputs[f"{frame}_Mzz"][:nelem] = forces.Txx[ic, 1::2]


class FloatingPost(om.ExplicitComponent):
    def initialize(self):
        self.options.declare("options")

    def setup(self):

        self.add_input("tower_elem_L", NULL * np.ones(MEMMAX), units="m")
        self.add_input("tower_elem_D", NULL * np.ones(MEMMAX), units="m")
        self.add_input("tower_elem_t", NULL * np.ones(MEMMAX), units="m")
        self.add_input("tower_elem_A", NULL * np.ones(MEMMAX), units="m**2")
        self.add_input("tower_elem_Asx", NULL * np.ones(MEMMAX), units="m**2")
        self.add_input("tower_elem_Asy", NULL * np.ones(MEMMAX), units="m**2")
        self.add_input("tower_elem_Ixx", NULL * np.ones(MEMMAX), units="kg*m**2")
        self.add_input("tower_elem_Iyy", NULL * np.ones(MEMMAX), units="kg*m**2")
        self.add_input("tower_elem_Izz", NULL * np.ones(MEMMAX), units="kg*m**2")
        self.add_input("tower_elem_E", NULL * np.ones(MEMMAX), units="Pa")
        self.add_input("tower_elem_G", NULL * np.ones(MEMMAX), units="Pa")
        self.add_input("tower_elem_sigma_y", NULL * np.ones(MEMMAX), units="Pa")
        self.add_input("tower_elem_qdyn", NULL * np.ones(MEMMAX), units="Pa")

        self.add_input("system_elem_L", NULL * np.ones(NELEM_MAX), units="m")
        self.add_input("system_elem_D", NULL * np.ones(NELEM_MAX), units="m")
        self.add_input("system_elem_t", NULL * np.ones(NELEM_MAX), units="m")
        self.add_input("system_elem_A", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_input("system_elem_Asx", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_input("system_elem_Asy", NULL * np.ones(NELEM_MAX), units="m**2")
        self.add_input("system_elem_Ixx", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_input("system_elem_Iyy", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_input("system_elem_Izz", NULL * np.ones(NELEM_MAX), units="kg*m**2")
        self.add_input("system_elem_E", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_input("system_elem_G", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_input("system_elem_sigma_y", NULL * np.ones(NELEM_MAX), units="Pa")
        self.add_input("system_elem_qdyn", NULL * np.ones(NELEM_MAX), units="Pa")

        # Processed Frame3DD/OpenFAST outputs
        self.add_input("system_Fz", NULL * np.ones(NELEM_MAX), units="N")
        self.add_input("system_Vx", NULL * np.ones(NELEM_MAX), units="N")
        self.add_input("system_Vy", NULL * np.ones(NELEM_MAX), units="N")
        self.add_input("system_Mxx", NULL * np.ones(NELEM_MAX), units="N*m")
        self.add_input("system_Myy", NULL * np.ones(NELEM_MAX), units="N*m")
        self.add_input("system_Mzz", NULL * np.ones(NELEM_MAX), units="N*m")

        self.add_input("tower_Fz", NULL * np.ones(NELEM_MAX), units="N")
        self.add_input("tower_Vx", NULL * np.ones(NELEM_MAX), units="N")
        self.add_input("tower_Vy", NULL * np.ones(NELEM_MAX), units="N")
        self.add_input("tower_Mxx", NULL * np.ones(NELEM_MAX), units="N*m")
        self.add_input("tower_Myy", NULL * np.ones(NELEM_MAX), units="N*m")
        self.add_input("tower_Mzz", NULL * np.ones(NELEM_MAX), units="N*m")

        self.add_output("constr_tower_stress", NULL * np.ones(NELEM_MAX))
        self.add_output("constr_tower_shell_buckling", NULL * np.ones(NELEM_MAX))
        self.add_output("constr_tower_global_buckling", NULL * np.ones(NELEM_MAX))

        self.add_output("constr_system_stress", NULL * np.ones(NELEM_MAX))
        self.add_output("constr_system_shell_buckling", NULL * np.ones(NELEM_MAX))
        self.add_output("constr_system_global_buckling", NULL * np.ones(NELEM_MAX))

    def compute(self, inputs, outputs):
        # Loop over tower and system analysis
        for frame in ["tower", "system"]:
            # Unpack some variables
            d = inputs[f"{frame}_elem_D"]
            nelem = np.where(d == NULL)[0][0]
            d = d[:nelem]
            t = inputs[f"{frame}_elem_t"][:nelem]
            h = inputs[f"{frame}_elem_L"][:nelem]
            Az = inputs[f"{frame}_elem_A"][:nelem]
            Asx = inputs[f"{frame}_elem_Asx"][:nelem]
            Jz = inputs[f"{frame}_elem_Izz"][:nelem]
            Iyy = inputs[f"{frame}_elem_Iyy"][:nelem]
            sigy = inputs[f"{frame}_elem_sigma_y"][:nelem]
            E = inputs[f"{frame}_elem_E"][:nelem]
            G = inputs[f"{frame}_elem_G"][:nelem]
            qdyn = inputs[f"{frame}_elem_qdyn"][:nelem]
            r = 0.5 * d

            gamma_f = self.options["options"]["gamma_f"]
            gamma_m = self.options["options"]["gamma_m"]
            gamma_n = self.options["options"]["gamma_n"]
            gamma_b = self.options["options"]["gamma_b"]

            # Get loads from Framee3dd/OpenFAST
            Fz = inputs[f"{frame}_Fz"][:nelem]
            Vx = inputs[f"{frame}_Vx"][:nelem]
            Vy = inputs[f"{frame}_Vy"][:nelem]
            Mxx = inputs[f"{frame}_Mxx"][:nelem]
            Myy = inputs[f"{frame}_Myy"][:nelem]
            Mzz = inputs[f"{frame}_Mzz"][:nelem]

            M = np.sqrt(Mxx ** 2 + Myy ** 2)
            V = np.sqrt(Vx ** 2 + Vy ** 2)

            # Initialize outputs
            outputs[f"constr_{frame}_stress"] = NULL * np.ones(NELEM_MAX)
            outputs[f"constr_{frame}_shell_buckling"] = NULL * np.ones(NELEM_MAX)
            outputs[f"constr_{frame}_global_buckling"] = NULL * np.ones(NELEM_MAX)

            # See http://svn.code.sourceforge.net/p/frame3dd/code/trunk/doc/Frame3DD-manual.html#structuralmodeling
            # print(Fz.shape, Az.shape, M.shape, r.shape, Iyy.shape)
            axial_stress = Fz / Az + M * r / Iyy
            shear_stress = np.abs(Mzz) / Jz * r + V / Asx
            hoop_stress = util_con.hoopStress(d, t, qdyn)
            outputs[f"constr_{frame}_stress"][:nelem] = util_con.vonMisesStressUtilization(
                axial_stress, hoop_stress, shear_stress, gamma_f * gamma_m * gamma_n, sigy
            )

            # Use DNV-GL CP202 Method
            check = util_dnvgl.CylinderBuckling(h, d, t, E=E, G=G, sigma_y=sigy, gamma=gamma_f * gamma_b)
            results = check.run_buckling_checks(Fz, M, axial_stress, hoop_stress, shear_stress)

            outputs[f"constr_{frame}_shell_buckling"][:nelem] = results["Shell"]
            outputs[f"constr_{frame}_global_buckling"][:nelem] = results["Global"]


class FloatingFrame(om.Group):
    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        opt = self.options["modeling_options"]

        self.add_subsystem("plat", PlatformFrame(options=opt), promotes=["*"])
        self.add_subsystem("pre", TowerPreMember(), promotes=["*"])

        prom = [
            "E_mat",
            "G_mat",
            "sigma_y_mat",
            "sigma_ult_mat",
            "wohler_exp_mat",
            "wohler_A_mat",
            "rho_mat",
            "rho_water",
            "unit_cost_mat",
            "material_names",
            "painting_cost_rate",
            "labor_cost_rate",
        ]
        prom += [
            ("nodes_xyz", "tower_nodes"),
            ("nodes_r", "tower_Rnode"),
            ("total_mass", "tower_mass"),
            ("total_cost", "tower_cost"),
            ("center_of_mass", "tower_center_of_mass"),
            ("joint1", "transition_node"),
            ("joint2", "tower_top_node"),
            ("Px", "tower_elem_Px"),
            ("Py", "tower_elem_Py"),
            ("Pz", "tower_elem_Pz"),
            ("qdyn", "tower_elem_qdyn"),
        ]
        for var in ["D", "t", "A", "Asx", "Asy", "rho", "Ixx", "Iyy", "Izz", "E", "G", "sigma_y"]:
            prom += [("section_" + var, "tower_elem_" + var)]

        prom += [
            "Uref",
            "zref",
            "z0",
            "shearExp",
            "cd_usr",
            "cm",
            "beta_wind",
            "rho_air",
            "mu_air",
            "beta_wave",
            "mu_water",
            "Uc",
            "Hsig_wave",
            "Tsig_wave",
            "water_depth",
        ]
        self.add_subsystem(
            "tower",
            Member(column_options=opt["floating"]["tower"], idx=0, n_mat=opt["materials"]["n_mat"]),
            promotes=prom,
        )
        self.add_subsystem("mux", PlatformTowerFrame(options=opt), promotes=["*"])
        self.add_subsystem("frame", FrameAnalysis(options=opt), promotes=["*"])
        self.add_subsystem("post", FloatingPost(options=opt["WISDEM"]["FloatingSE"]), promotes=["*"])



"""Common processes and cargo types for Cable Installations."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from marmot import process
from wisdem.orbit.core.logic import position_onsite
from wisdem.orbit.core.defaults import process_times as pt


class SimpleCable:
    """Simple Cable Class"""

    def __init__(self, linear_density):
        """
        Creates an instance of SimpleCable.

        Parameters
        ----------
        linear_density : int | float
        """

        self.linear_density = linear_density


@process
def load_cable_on_vessel(vessel, cable, constraints={}, **kwargs):
    """
    Subprocess for loading `cable` onto the configured `vessel`.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Required to have configured `cable_storage`.
    cable : SimpleCable | Cable
        Cable type.
    constraints : dict
        Constraints to be applied to cable loading subprocess.
    """

    key = "cable_load_time"
    load_time = kwargs.get(key, pt[key])

    vessel.cable_storage.load_cable(cable)
    yield vessel.task_wrapper("Load Cable", load_time, constraints=constraints, **kwargs)


@process
def landfall_tasks(vessel, trench_length, **kwargs):
    """
    List of tasks that must be completed at landfall at the beginning of the
    export system installation process.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Required to have configured `cable_storage`.
    trench_length : int | float
        Length of trench that is dug through the beach (km).
    """

    yield tow_plow(vessel, trench_length, **kwargs)
    yield pull_winch(vessel, trench_length, **kwargs)
    yield prep_cable(vessel, **kwargs)
    yield pull_in_cable(vessel, **kwargs)
    yield terminate_cable(vessel, **kwargs)
    yield lower_cable(vessel, **kwargs)


@process
def prep_cable(vessel, **kwargs):
    """
    Task representing time required to prepare cable for pull-in.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Requires configured `transit_limits`.
    """

    key = "cable_prep_time"
    prep_time = kwargs.get(key, pt[key])

    yield vessel.task_wrapper("Prepare Cable", prep_time, constraints=vessel.transit_limits, **kwargs)


@process
def lower_cable(vessel, **kwargs):
    """
    Task representing time required to lower cable to seafloor.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Requires configured `operational_limits`.
    """

    key = "cable_lower_time"
    lower_time = kwargs.get(key, pt[key])

    yield vessel.task_wrapper(
        "Lower Cable",
        lower_time,
        constraints=vessel.operational_limits,
        **kwargs,
    )


@process
def pull_in_cable(vessel, **kwargs):
    """
    Task representing time required to pull cable into offshore substructure or
    at onshore trench.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Requires configured `operational_limits`.
    """

    key = "cable_pull_in_time"
    pull_in_time = kwargs.get(key, pt[key])

    yield vessel.task_wrapper(
        "Pull In Cable",
        pull_in_time,
        constraints=vessel.operational_limits,
        **kwargs,
    )


@process
def terminate_cable(vessel, **kwargs):
    """
    Task representing time required to terminate and test cable connection.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Requires configured `operational_limits`.
    """

    key = "cable_termination_time"
    termination_time = kwargs.get(key, pt[key])

    yield vessel.task_wrapper(
        "Terminate Cable",
        termination_time,
        constraints=vessel.operational_limits,
        **kwargs,
    )


@process
def lay_bury_cable(vessel, distance, **kwargs):
    """
    Task representing time required to lay and bury a cable section.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Requires configured `operational_limits`.
    distance : int | float
        Distance of cable section (km)
    cable_lay_bury_speed : int | float
        Maximum speed at which cable is dispensed (km/hr)
    """

    kwargs = {**kwargs, **getattr(vessel, "_transport_specs", {})}

    key = "cable_lay_bury_speed"
    lay_bury_speed = kwargs.get(key, pt[key])
    lay_bury_time = distance / lay_bury_speed

    yield vessel.task_wrapper(
        "Lay/Bury Cable",
        lay_bury_time,
        constraints=vessel.operational_limits,
        suspendable=True,
        **kwargs,
    )


@process
def lay_cable(vessel, distance, **kwargs):
    """
    Task representing time required to lay a cable section.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Requires configured `operational_limits`.
    distance : int | float
        Distance of cable section (km).
    cable_lay_speed : int | float
        Maximum speed at which cable is dispensed (km/hr)
    """

    kwargs = {**kwargs, **getattr(vessel, "_transport_specs", {})}

    key = "cable_lay_speed"
    lay_speed = kwargs.get(key, pt[key])
    lay_time = distance / lay_speed

    yield vessel.task_wrapper(
        "Lay Cable",
        lay_time,
        constraints=vessel.operational_limits,
        suspendable=True,
        **kwargs,
    )


@process
def bury_cable(vessel, distance, **kwargs):
    """
    Task representing time required to bury a cable section.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Requires configured `operational_limits`.
    distance : int | float
        Distance of cable section (km).
    cable_bury_speed : int | float
        Maximum speed at which cable is buried (km/hr).
    """

    kwargs = {**kwargs, **getattr(vessel, "_transport_specs", {})}

    key = "cable_bury_speed"
    bury_speed = kwargs.get(key, pt[key])
    bury_time = distance / bury_speed

    yield vessel.task_wrapper(
        "Bury Cable",
        bury_time,
        constraints=vessel.operational_limits,
        suspendable=True,
        **kwargs,
    )


@process
def splice_cable(vessel, **kwargs):
    """
    Task representing time required to splice a cable at sea.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Requires configured `operational_limits`.
    cable_splice_time : int | float
        Time required to splice two cable ends together (h).
    """

    key = "cable_splice_time"
    splice_time = kwargs.get(key, pt[key])

    yield vessel.task_wrapper(
        "Splice Cable",
        splice_time,
        constraints=vessel.operational_limits,
        **kwargs,
    )


@process
def raise_cable(vessel, **kwargs):
    """
    Task representing time required to raise the unspliced cable from the
    seafloor.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Requires configured `operational_limits`.
    cable_raise_time : int | float
        Time required to raise the cable from the seafloor (h).
    """

    key = "cable_raise_time"
    raise_time = kwargs.get(key, pt[key])

    yield vessel.task_wrapper(
        "Raise Cable",
        raise_time,
        constraints=vessel.operational_limits,
        **kwargs,
    )


@process
def splice_process(vessel, **kwargs):
    """
    A list of tasks representing the entire cable splicing process.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Requires configured `operational_limits`.
    """

    yield position_onsite(vessel)
    yield raise_cable(vessel, **kwargs)
    yield splice_cable(vessel, **kwargs)
    yield lower_cable(vessel, **kwargs)


@process
def tow_plow(vessel, distance, **kwargs):
    """
    Task representing time required to tow plow at landfall site.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Requires configured `operational_limits`.
    distance : int | float
        Distance to between cable laying vessel and onshore construction (km).
    tow_plow_speed : float
        Towing speed (km/h).
    """

    key = "tow_plow_speed"
    plow_speed = kwargs.get(key, pt[key])
    plow_time = distance / plow_speed

    yield vessel.task_wrapper("Tow Plow", plow_time, constraints=vessel.operational_limits, **kwargs)


@process
def pull_winch(vessel, distance, **kwargs):
    """
    Task representing time required to pull cable onshore through the
    previously dug trench.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Requires configured `operation_limits`.
    distance: int | float
        Distance winch wire must travel to reach the plow.
    pull_winch_speed : int | float
        Speed at wich the winch travels (km/h).
    """

    key = "pull_winch_speed"
    winch_speed = kwargs.get(key, pt[key])
    winch_time = distance / winch_speed

    yield vessel.task_wrapper(
        "Pull Winch",
        winch_time,
        constraints=vessel.operational_limits,
        **kwargs,
    )


@process
def dig_trench(vessel, distance, **kwargs):
    """
    Task representing time required to dig a trench prior to cable lay and burial

    Parameters
    ----------
    vessel : Vessel
        Performing vessel. Requires configured `operational_limits`.
    distance : int | float
        Length of trench, equal to length of cable section (km).
    trench_dig_speed : int | float
        Speed at which trench is dug (km/hr).
    """

    kwargs = {**kwargs, **getattr(vessel, "_transport_specs", {})}

    key = "trench_dig_speed"
    trench_dig_speed = kwargs.get(key, pt[key])
    trench_dig_time = distance / trench_dig_speed

    yield vessel.task_wrapper(
        "Dig Trench",
        trench_dig_time,
        constraints=vessel.operational_limits,
        suspendable=True,
        **kwargs,
    )



"""
**CollectionCost.py**
- Created by Matt Shields for Offshore BOS
- Refactored by Parangat Bhaskar for LandBOSSE

NREL - 05/31/2019

This module consists of two classes:

- The first class in this module is the parent class Cable, with a sublass Array that inherits from Cable

- The second class is the ArraySystem class that instantiates the Array class and determines the wind farm layout and calculates total collection system cost
"""

import math
import traceback

import numpy as np
import pandas as pd
from wisdem.landbosse.model.CostModule import CostModule
from wisdem.landbosse.model.WeatherDelay import WeatherDelay as WD


class Cable:
    """

    Create an instance of Cable (either array or export)

        Parameters
        ---------
        cable_specs : dict
            Dictionary containing cable specifications
        line_frequency_hz : int
            Additional user inputs

        Returns
        -------
        current_capacity : float
            Cable current rating at 1m burial depth, Amps
        rated_voltage : float
            Cable rated voltage, kV
        ac_resistance : float
            Cable resistance for AC current, Ohms/km
        inductance : float
            Cable inductance, mH/km
        capacitance : float
            Cable capacitance, nF/km
        cost : int
            Cable cost, $US/km
        char_impedance : float
            Characteristic impedance of equivalent cable circuit, Ohms
        power_factor : float
            Power factor of AC current in cable (nondim)
        cable_power : float
            Maximum 3-phase power dissipated in cable, MW

    """

    def __init__(self, cable_specs, addl_specs):
        """
        Parameters
        ----------
        cable_specs : dict
            The input dictionary with key value pairs described in the
            class documentation

        addl_specs : dict
            The output dictionary with key value pairs as found on the
            output documentation.

        """

        self.current_capacity = cable_specs["Current Capacity (A)"]
        self.rated_voltage = cable_specs["Rated Voltage (V)"]
        self.ac_resistance = cable_specs["AC Resistance (Ohms/km)"]
        self.inductance = cable_specs["Inductance (mH/km)"]
        self.capacitance = cable_specs["Capacitance (nF/km)"]
        self.cost = cable_specs["Cost (USD/LF)"]
        self.line_frequency_hz = addl_specs["line_frequency_hz"]

        # Calc additional cable specs
        self.calc_char_impedance(self.line_frequency_hz)
        self.calc_power_factor()
        self.calc_cable_power()

    def calc_char_impedance(self, line_frequency_hz):
        """
        Calculate characteristic impedance of cable, Ohms

        Parameters
        ----------
        line_frequency_hz : int
            Frequency of AC current, Hz
        """
        conductance = 1 / self.ac_resistance

        num = complex(self.ac_resistance, 2 * math.pi * line_frequency_hz * self.inductance)
        den = complex(conductance, 2 * math.pi * line_frequency_hz * self.capacitance)
        self.char_impedance = np.sqrt(num / den)

    def calc_power_factor(self):
        """
        Calculate power factor
        """

        phase_angle = math.atan(np.imag(self.char_impedance) / np.real(self.char_impedance))
        self.power_factor = math.cos(phase_angle)

    def calc_cable_power(self):
        """
        Calculate maximum power transfer through 3-phase cable, MW
        """

        # TODO: Verify eqn is correct
        self.cable_power = np.sqrt(3) * self.rated_voltage * self.current_capacity * self.power_factor / 1000


class Array(Cable):
    """Array cable base class"""

    def __init__(self, cable_specs, addl_inputs):
        """
        Creates an instance of Array cable.
        (May be multiple instances of different capacity cables in a string)

        Parameters
        ----------
        cable_specs : dict
            Dictionary containing following cable specifications:

            - turbine_rating_MW

            - upstream_turb

            - turbine_spacing_rotor_diameters

            - rotor_diameter_m

        addl_inputs : dict

            - Any additional user inputs

        Returns
        -------
        self.max_turb_per_cable : float
            Maximum number of turbines (at turbine_rating_MW) an individual cable
            can support
        self.num_turb_per_cable : float
            Number of turbines each cable in a string actually supports.
        self.turb_sequence : float
            Ordering of cable in string, starting with smallest cable at 0
        self.downstream_connection : int
            Additional cable length requried to connect between different sized
            cables (for first cable in string only)
        self.array_cable_len : float
            Length of individual cable in a string, km
        """

        super().__init__(cable_specs, addl_inputs)
        self.line_frequency_hz = addl_inputs["line_frequency_hz"]
        self.calc_max_turb_per_cable(addl_inputs)
        self.calc_num_turb_per_cable(addl_inputs)
        self.calc_array_cable_len(addl_inputs)

    def calc_max_turb_per_cable(self, addl_inputs):
        """
        Calculate the number of turbines that each cable can support

        Parameters
        ----------
        turbine_rating_MW : int
            Nameplate capacity of individual turbines
        """

        turbine_rating_MW = addl_inputs["turbine_rating_MW"]

        self.max_turb_per_cable = np.floor(self.cable_power / turbine_rating_MW)

    def calc_num_turb_per_cable(self, addl_inputs):
        """
        Calculates actual number of turbines per cable, accounting for upstream
        turbines.

        Parameters
        ----------
        upstream_turb : int
            Number of turbines on upstream cables in string
        """

        upstream_turb = addl_inputs["upstream_turb"]
        self.turb_sequence = addl_inputs["turb_sequence"]

        self.num_turb_per_cable = self.max_turb_per_cable - upstream_turb

        if upstream_turb == 0:
            self.downstream_connection = -1
        else:
            self.downstream_connection = 0

    def calc_array_cable_len(self, addl_inputs):
        """
        Calculate array cable length per string, km

        Parameters
        ----------
        turbine_spacing_rotor_diameters : int
            Spacing between turbines in string, # of rotor diameters
        rotor_diameter_m : int or float
            Rotor diameter, m
        """

        turbine_spacing_rotor_diameters = addl_inputs["turbine_spacing_rotor_diameters"]
        rotor_diameter_m = addl_inputs["rotor_diameter_m"]

        self.calc_turb_section_len(turbine_spacing_rotor_diameters, rotor_diameter_m)

        self.array_cable_len = (self.num_turb_per_cable + self.downstream_connection) * self.turb_section_length

    #    @staticmethod
    def calc_turb_section_len(self, turbine_spacing_rotor_diameters, rotor_diameter_m):
        """
        Calculate array cable section length between two turbines. Also, section length == trench length. Which means
        trench_length = cable_length for that section.

        Parameters
        ----------
        turbine_spacing_rotor_diameters : int
            Spacing between turbines in string, # of rotor diameters
        rotor_diameter_m : int or float
            Rotor diameter, m

        Returns
        -------
        turb_connect_len : int
            Length of array cable between two turbines, km
        """

        self.turb_section_length = (turbine_spacing_rotor_diameters * rotor_diameter_m) / 1000

        return self.turb_section_length


class ArraySystem(CostModule):
    """


    \nThis module:

    * Calculates cable length to substation

    * Calculates number of strings in a subarray

    * Calculated number of strings

    * Calculates total cable length for each cable type

    * Calculates total trench length

    * Calculates total collection system cost based on amount of material, amount of labor, price data, cable length, and trench length.



    **Keys in the input dictionary are the following:**

    * Given below are attributes that define each cable type:
        * conductor_size
            (int)   cross-sectional diameter of cable [in mm]



    """

    def __init__(self, input_dict, output_dict, project_name):

        self.input_dict = input_dict
        self.output_dict = output_dict
        self.project_name = project_name
        self.output_dict["total_cable_len_km"] = 0
        self._km_to_LF = 0.0003048  # Units: [km/LF] Conversion factor for converting from km to linear foot.
        self._total_cable_cost = 0
        self._total_turbine_counter = 0
        self.turbines_on_cable = []
        self._cable_length_km = dict()
        self.check_terminal = 0

    def calc_num_strings(self):
        """
        Calculate number of full and partial strings to support full plant
        capacity.

        Parameters
        ----------
        available cables : dict
            Dictionary of cable types
        plant_capacity : int | float
            Total capcity of wind plant (MW)
        turbine_capacity : int | float
            Nameplate capacity of individual turbines (MW)

        Returns
        -------
        self.output_dict['total_turb_per_string'] : float
            Number of turbines on each string
        self.output_dict['num_full_strings'] : float
            Number of complete strings in array
        turb_per_partial_string : float
            Number of turbines in the partial string (if applicable)
        self.output_dict['num_partial_strings'] : float
            Number of partial strings (if applicable, 0 or 1)
        perc_full_string : list
            Percentage of maximum number of turbines per cable type on
            partial string
        self.output_dict['num_turb_per_cable'] : list
            Number of turbines on each cable type in string
        """

        # Calculate total number of individual turbines in wind plant
        self.output_dict["total_turb"] = self.input_dict["num_turbines"]

        # Calculate the number of turbines on each cable type in a string
        self.output_dict["num_turb_per_cable"] = [cable.num_turb_per_cable for cable in self.cables.values()]

        # Calculate the total number of turbines per string
        self.output_dict["total_turb_per_string"] = sum(self.output_dict["num_turb_per_cable"])

        # Calculate number of full strings and any remainder required to
        # support the total number of turbines
        self.output_dict["num_full_strings"] = np.floor(
            self.output_dict["total_turb"] / self.output_dict["total_turb_per_string"]
        )
        self.output_dict["num_leftover_turb"] = (
            self.output_dict["total_turb"] % self.output_dict["total_turb_per_string"]
        )

        # Calculate number of turbines on a remaining partial string

        # Note: self.output_dict['turb_per_partial_string'] is only set if
        # calc_num_turb_partial_strings()
        # is called, which isn't always the case, as seen in the if...else construct below
        #
        # This means that self.output_dict['turb_per_partial_string'] cannot
        # be used an output value for the details output.

        if self.output_dict["num_leftover_turb"] > 0:
            self.output_dict["num_partial_strings"] = 1
            self.output_dict["perc_partial_string"] = self.calc_num_turb_partial_strings(
                self.output_dict["num_leftover_turb"], self.output_dict["num_turb_per_cable"]
            )
        else:
            self.output_dict["num_partial_strings"] = 0
            self.output_dict["perc_partial_string"] = np.zeros(len(self.output_dict["num_turb_per_cable"]))

        return (
            self.output_dict["total_turb_per_string"],
            self.output_dict["num_full_strings"],
            self.output_dict["num_partial_strings"],
            self.output_dict["perc_partial_string"],
            self.output_dict["num_turb_per_cable"],
        )

    def calc_num_turb_partial_strings(self, num_leftover_turb, num_turb_per_cable):
        """
        If a partial string exists, calculate the percentage of turbines on
        each cable relative to a full string

        Parameters
        ----------
        self.output_dict['num_leftover_turb'] : float
            Number of turbines in partial string
        self.output_dict['num_turb_per_cable'] : list
            List of number of turbines per cable type on a full string

        Returns
        -------
        np.array
            Array of percent of turbines per cable type on partial string
            relative to full string
        """

        num_remaining = num_leftover_turb
        turb_per_partial_string = []

        # Loop through each cable type in the string. Determine how many
        # turbines are required for each cable type on the partial string
        for max_turb in num_turb_per_cable:
            if num_remaining > 0:
                turb_per_partial_string.append(min(num_remaining, max_turb))
            else:
                turb_per_partial_string.append(0.0)
            num_remaining -= max_turb

        perc_partial_string = np.divide(turb_per_partial_string, num_turb_per_cable)

        # Check to make sure there aren't any zeros in num_turbines_per_cable, which is used as the denominator
        # in the division above (this happens when not all of the cable types in the input sheet need to be used).
        # If there is a zero, then print a warning and change NaN to 0 in perc_partial_string.
        if 0.0 in num_turb_per_cable:
            print(
                f"Warning: {self.project_name} CollectionCost module generates number of turbines per string that "
                f"includes a zero entry. Please confirm that there not all cable types need to be used for the number of turbines that are being run."
                f' num_turbines={self.input_dict["num_turbines"]} rating_MW={self.input_dict["turbine_rating_MW"]}'
                f" num_turb_per_cable: {num_turb_per_cable}"
            )
            perc_partial_string = np.nan_to_num(perc_partial_string)

        self.output_dict["turb_per_partial_string"] = turb_per_partial_string

        return perc_partial_string

    # TODO: change length_to_substation calculation as a user defined input?
    @staticmethod
    def calc_cable_len_to_substation(
        distance_to_grid, turbine_spacing_rotor_diameters, row_spacing_rotor_diameters, num_strings
    ):
        """
        Calculate the distance for the largest cable run to substation
        Assumes substation is in the center of the layout, 1 row spacing in
        front of first row

        Parameters
        ----------
        turbine_spacing_rotor_diameters : int or float
            Spacing between turbines in a row, # of rotor diameters
        row_spacing_rotor_diameters : int or float
            Spacing between rows in wind plant, # of rotor diameters
        num_strings : int
            Total number of strings

        Returns
        -------
        len_to_substation : int or float
            Total length of largest array cable required to connect each string
            to substation, km
        """

        string_to_substation_length = []

        if num_strings > 1:
            # Define spacing terms for even or odd number of strings
            #   Even number: substation centered between middle two strings
            #   Odd number : substation centered on middle string
            if (num_strings % 2) == 0:
                n_max = int(num_strings / 2)
                turb_space_scaling = 0.5
                range_strings = range(1, n_max + 1)

            else:
                n_max = int((num_strings - 1) / 2)
                turb_space_scaling = 1
                range_strings = range(n_max + 1)

            # Calculate hypotenuse length of each string to substation
            for idx in range_strings:
                if idx == 0:
                    c = 1
                else:
                    c = 2
                string_to_substation_length.append(
                    c
                    * np.sqrt(
                        row_spacing_rotor_diameters ** 2
                        + (turb_space_scaling * idx * turbine_spacing_rotor_diameters) ** 2
                    )
                )

        else:
            string_to_substation_length.append(distance_to_grid)

        # Sum up total length to substation
        len_to_substation = np.sum(string_to_substation_length)

        return len_to_substation

    # TODO: Add parameter info in docstrings
    @staticmethod
    def calc_total_cable_length(
        total_turbines,
        count,
        check_terminal,
        turbines_per_cable,
        cable,
        cable_specs,
        num_full_strings,
        num_partial_strings,
        len_to_substation,
        perc_partial_string,
    ):
        """
        Calculate total length of each cable type, km

        Parameters
        ----------
        cable : object
            Instance of individual cable type
        cable_specs : dict
            Dictionary containing cable specifications
        self.output_dict['num_full_strings'] : float
            Number of complete strings in array
        self.output_dict['num_partial_strings'] : float
            Number of partial strings (if applicable, 0 or 1)
        len_to_substation : int or float
            Total length of largest array cable required to connect each string
            to substation, km
        self.output_dict['perc_partial_string'] : list
            List of percent of turbines per cable type on partial string
            relative to full string

        Returns
        -------
        total_cable_len : int or float
            Total length of individual cable type
        """

        # If terminal cable has already been accounted for, skip any
        # calculations for other cables.
        if (cable.turb_sequence - 1) > check_terminal:
            cable.array_cable_len = 0
            cable.total_length = 0
            cable.num_turb_per_cable = 0
            return 0, 0

        # If num full strings < = 1, find which cable the final turbine
        # is on, and calculate total cable length (including the len to
        # substation) using that cable.

        # This 'elif' is essentially a switch for distributed wind:
        elif num_full_strings < 1 and num_partial_strings >= 0:

            # if number of turbines is less than total string capacity,
            # find the terminal cable and find total cable len
            # up till that cable.

            # If total turbines in project are less than cumulative turbines
            # up till and including that cable.

            terminal_string = cable.turb_sequence - 1  # Flag this cable as it is
            # also the actual terminal cable

            if (cable.turb_sequence - 1) == 0:  # That is, if cable # 1 can hold
                # more turbines than specified by user, it is the terminal cable

                cable.num_turb_per_cable = total_turbines
                cable.array_cable_len = (
                    cable.num_turb_per_cable + cable.downstream_connection
                ) * cable.turb_section_length

                total_cable_len = (
                    (num_full_strings * cable.array_cable_len) + (num_partial_strings * cable.array_cable_len)
                ) + len_to_substation

            else:

                cable.num_turb_per_cable = total_turbines - turbines_per_cable[(count - 1)]
                cable.array_cable_len = (
                    cable.num_turb_per_cable + cable.downstream_connection
                ) * cable.turb_section_length

                total_cable_len = (
                    (num_full_strings * cable.array_cable_len) + (num_partial_strings * cable.array_cable_len)
                ) + len_to_substation

            return total_cable_len, terminal_string

        else:  # Switch for utility scale landbosse
            if cable.turb_sequence == len(cable_specs):

                # Only add len_to_substation to the final cable in the string
                total_cable_len = (
                    num_full_strings * cable.array_cable_len
                    + num_partial_strings * (cable.array_cable_len * perc_partial_string)
                    + len_to_substation
                )
            else:
                total_cable_len = num_full_strings * cable.array_cable_len + num_partial_strings * (
                    cable.array_cable_len * perc_partial_string
                )

        # here 9999 == flag to announce that the terminal cable has NOT been reached
        # and to continue calculations for each cable
        return total_cable_len, 9999

    def create_ArraySystem(self):
        # data used in parent classes:
        self.addl_specs = dict()
        self.addl_specs["turbine_rating_MW"] = self.input_dict["turbine_rating_MW"]
        self.addl_specs["upstream_turb"] = 0
        self.addl_specs["turb_sequence"] = 1
        self.addl_specs["turbine_spacing_rotor_diameters"] = self.input_dict["turbine_spacing_rotor_diameters"]
        self.addl_specs["rotor_diameter_m"] = self.input_dict["rotor_diameter_m"]
        self.addl_specs["line_frequency_hz"] = self.input_dict["line_frequency_hz"]

        system = {
            "upstream_turb": self.addl_specs["upstream_turb"],
            "turb_sequence": self.addl_specs["turb_sequence"],
            "turbine_rating_MW": self.addl_specs["turbine_rating_MW"],
            "turbine_spacing_rotor_diameters": self.addl_specs["turbine_spacing_rotor_diameters"],
            "rotor_diameter_m": self.addl_specs["rotor_diameter_m"],
        }

        # Loops through all user defined array cable types, composing them
        # in ArraySystem

        self.cables = {}
        self.input_dict["cable_specs"] = self.input_dict["cable_specs_pd"].T.to_dict()
        n = 0  # to keep tab of number of cables input by user.
        while n < len(self.input_dict["cable_specs"]):
            specs = self.input_dict["cable_specs"][n]
            # Create instance of each cable and assign to ArraySystem.cables
            cable = Array(specs, self.addl_specs)
            n += 1

            # self.cables[name] stores value which is a new instantiation of object of type Array.
            self.cables[specs["Array Cable"]] = cable
            self.output_dict["cables"] = self.cables

            # Update number of upstream cables on the string
            self.addl_specs["upstream_turb"] += cable.num_turb_per_cable
            self.addl_specs["turb_sequence"] += 1

        # Calculate number of required strings to support plant capacity
        (
            self.output_dict["turb_per_string"],
            self.output_dict["num_full_strings"],
            self.output_dict["num_partial_strings"],
            self.output_dict["perc_partial_string"],
            self.output_dict["num_turb_per_cable"],
        ) = self.calc_num_strings()

        # Calculate total length of cable run to substation
        self.output_dict["num_strings"] = self.output_dict["num_full_strings"] + self.output_dict["num_partial_strings"]

        if self.input_dict["user_defined_distance_to_grid_connection"] == 0:  # where (0 = No) and (1 = Yes)

            # This only gets used if number of strings is <= 1 :
            distributed_wind_distance_to_grid = (
                self.input_dict["turbine_spacing_rotor_diameters"] * self.input_dict["rotor_diameter_m"]
            ) / 1000
            self.output_dict["distance_to_grid_connection_km"] = self.calc_cable_len_to_substation(
                distributed_wind_distance_to_grid,
                self.input_dict["turbine_spacing_rotor_diameters"],
                self.input_dict["row_spacing_rotor_diameters"],
                self.output_dict["num_strings"],
            )
        else:
            self.output_dict["distance_to_grid_connection_km"] = self.input_dict["distance_to_grid_connection_km"]

        self.output_dict["cable_len_to_grid_connection_km"] = self.output_dict[
            "distance_to_grid_connection_km"
        ]  # assumes 3 conductors and fiber and neutral

        cable_sequence = 0
        # Make a list of how many turbines per cable
        for _, (name, cable) in enumerate(self.cables.items()):
            if cable_sequence == 0:
                self.turbines_on_cable.append(cable.num_turb_per_cable)
            else:
                self.turbines_on_cable.append(cable.num_turb_per_cable + self.turbines_on_cable[(cable_sequence - 1)])
            # turbines_on_cable[cable_sequence] += cable.num_turb_per_cable
            cable_sequence += 1
        self.__turbines_on_cable = self.turbines_on_cable

        # Calculate total length of each cable type, and total cost that calculated length of cable:
        count = 0
        for idx, (name, cable) in enumerate(self.cables.items()):
            total_cable_len, self.check_terminal = self.calc_total_cable_length(
                self.output_dict["total_turb"],
                count,
                self.check_terminal,
                self.__turbines_on_cable,
                cable,
                self.input_dict["cable_specs"],
                self.output_dict["num_full_strings"],
                self.output_dict["num_partial_strings"],
                self.output_dict["distance_to_grid_connection_km"],
                self.output_dict["perc_partial_string"][idx],
            )
            count += 1
            # self._total_turbine_counter = turbine_tally
            self._cable_length_km[name] = total_cable_len
            cable.total_length = total_cable_len
            self.output_dict["total_cable_len_km"] += total_cable_len
            cable.total_cost = (total_cable_len / self._km_to_LF) * cable.cost
            self._total_cable_cost += cable.total_cost  # Keep running tally of total cable cost used in wind farm.

        # Repopulate the turbines per cable sequence to make sure it reflects any changes that happened since
        # the first time this sequence was populated.
        self.output_dict["num_turb_per_cable"] = [cable.num_turb_per_cable for cable in self.cables.values()]
        self.output_dict["total_turb_per_string"] = sum(self.output_dict["num_turb_per_cable"])

    def calculate_trench_properties(self, trench_properties_input, trench_properties_output):
        """
        Calculates the length of trench needed based on cable length and width of mulcher.
        """

        # units of cubic meters
        trench_properties_output["trench_length_km"] = trench_properties_output["total_cable_len_km"]

    def calculate_weather_delay(self, weather_delay_input_data, weather_delay_output_data):
        """Calculates wind delays for roads"""
        # construct WeatherDelay module
        WD(weather_delay_input_data, weather_delay_output_data)

        # compute weather delay
        wind_delay = pd.DataFrame(weather_delay_output_data["wind_delays"])

        # if greater than 4 hour delay, then shut down for full day (10 hours)
        wind_delay[(wind_delay > 4)] = 10
        weather_delay_output_data["wind_delay_time"] = float(wind_delay.sum())

        return weather_delay_output_data

    def estimate_construction_time(self, construction_time_input_data, construction_time_output_data):
        """
        Function to estimate construction time on per turbine basis.

        Parameters
        -------
        duration_construction

        pd.DataFrame
            rsmeans

        pd.DataFrame
            trench_length_km



        Returns
        -------

        (pd.DataFrame) operation_data

        """

        collection_construction_time = (
            construction_time_input_data["construct_duration"] * 1 / 3
        )  # assumes collection construction occurs for one-third of project duration

        throughput_operations = construction_time_input_data["rsmeans"]
        trench_length_km = construction_time_output_data["trench_length_km"]
        if construction_time_input_data["turbine_rating_MW"] >= 0.1:
            operation_data = throughput_operations.where(throughput_operations["Module"] == "Collection").dropna(
                thresh=4
            )
            # from rsmeans data, only read in Collection related data and filter out the rest:
            cable_trenching = throughput_operations[throughput_operations.Module == "Collection"]
        else:  # switch for small DW
            operation_data = throughput_operations.where(
                throughput_operations["Module"] == "Small DW Collection"
            ).dropna(thresh=4)
            # from rsmeans data, only read in Collection related data and filter out the rest:
            cable_trenching = throughput_operations[throughput_operations.Module == "Small DW Collection"]
        # operation_data = pd.merge()

        # from rsmeans data, only read in Collection related data and filter out the rest:
        cable_trenching = throughput_operations[throughput_operations.Module == "Collection"]

        # Storing data with labor related inputs:
        trenching_labor = cable_trenching[cable_trenching.values == "Labor"]
        trenching_labor_usd_per_hr = trenching_labor["Rate USD per unit"].sum()

        construction_time_output_data["trenching_labor_usd_per_hr"] = trenching_labor_usd_per_hr
        trenching_labor_daily_output = trenching_labor["Daily output"].values[
            0
        ]  # Units:  LF/day  -> where LF = Linear Foot
        trenching_labor_num_workers = trenching_labor["Number of workers"].sum()

        # Storing data with equipment related inputs:
        trenching_equipment = cable_trenching[cable_trenching.values == "Equipment"]
        trenching_cable_equipment_usd_per_hr = trenching_equipment["Rate USD per unit"].sum()
        construction_time_output_data["trenching_cable_equipment_usd_per_hr"] = trenching_cable_equipment_usd_per_hr
        trenching_equipment_daily_output = trenching_equipment["Daily output"].values[
            0
        ]  # Units:  LF/day  -> where LF = Linear Foot
        construction_time_output_data["trenching_labor_daily_output"] = trenching_labor_daily_output
        construction_time_output_data["trenching_equipment_daily_output"] = trenching_equipment_daily_output

        operation_data["Number of days taken by single crew"] = (
            trench_length_km / self._km_to_LF
        ) / trenching_labor_daily_output
        operation_data["Number of crews"] = np.ceil(
            (operation_data["Number of days taken by single crew"] / 30) / collection_construction_time
        )
        operation_data["Cost USD without weather delays"] = (
            (trench_length_km / self._km_to_LF) / trenching_labor_daily_output
        ) * (operation_data["Rate USD per unit"] * construction_time_input_data["operational_hrs_per_day"])
        alpha = operation_data[operation_data["Type of cost"] == "Collection"]
        operation_data_id_days_crews_workers = alpha[
            ["Operation ID", "Number of days taken by single crew", "Number of crews", "Number of workers"]
        ]

        alpha = operation_data[operation_data["Type of cost"] == "Labor"]
        operation_data_id_days_crews_workers = alpha[
            ["Operation ID", "Number of days taken by single crew", "Number of crews", "Number of workers"]
        ]

        # if more than one crew needed to complete within construction duration then assume that all construction
        # happens within that window and use that timeframe for weather delays;
        # if not, use the number of days calculated
        operation_data["time_construct_bool"] = (
            operation_data["Number of days taken by single crew"] > collection_construction_time * 30
        )
        boolean_dictionary = {True: collection_construction_time * 30, False: np.NAN}
        operation_data["time_construct_bool"] = operation_data["time_construct_bool"].map(boolean_dictionary)
        operation_data["Time construct days"] = operation_data[
            ["time_construct_bool", "Number of days taken by single crew"]
        ].min(axis=1)
        num_days = operation_data["Time construct days"].max()

        # No 'management crew' in small DW
        if construction_time_input_data["turbine_rating_MW"] >= 0.1:
            # pull out management data
            crew_cost = self.input_dict["crew_cost"]
            crew = self.input_dict["crew"][self.input_dict["crew"]["Crew type ID"].str.contains("M0")]
            management_crew = pd.merge(crew_cost, crew, on=["Labor type ID"])
            management_crew = management_crew.assign(
                per_diem_total=management_crew["Per diem USD per day"] * management_crew["Number of workers"] * num_days
            )
            management_crew = management_crew.assign(
                hourly_costs_total=management_crew["Hourly rate USD per hour"]
                * self.input_dict["hour_day"][self.input_dict["time_construct"]]
                * num_days
            )
            management_crew = management_crew.assign(
                total_crew_cost_before_wind_delay=management_crew["per_diem_total"]
                + management_crew["hourly_costs_total"]
            )
            self.output_dict["management_crew"] = management_crew
            self.output_dict["managament_crew_cost_before_wind_delay"] = management_crew[
                "total_crew_cost_before_wind_delay"
            ].sum()
        else:
            self.output_dict["managament_crew_cost_before_wind_delay"] = 0.0

        construction_time_output_data["operation_data_id_days_crews_workers"] = operation_data_id_days_crews_workers
        construction_time_output_data["operation_data_entire_farm"] = operation_data

        return construction_time_output_data["operation_data_entire_farm"]

    def calculate_costs(self, calculate_costs_input_dict, calculate_costs_output_dict):

        # read in rsmeans data:
        # rsmeans = calculate_costs_input_dict['rsmeans']
        operation_data = calculate_costs_output_dict["operation_data_entire_farm"]

        per_diem = (
            operation_data["Number of workers"]
            * operation_data["Number of crews"]
            * (operation_data["Time construct days"] + np.ceil(operation_data["Time construct days"] / 7))
            * calculate_costs_input_dict["rsmeans_per_diem"]
        )
        per_diem = per_diem.dropna()

        calculate_costs_output_dict["time_construct_days"] = (
            calculate_costs_output_dict["trench_length_km"] / self._km_to_LF
        ) / calculate_costs_output_dict["trenching_labor_daily_output"]
        wind_delay_fraction = (
            calculate_costs_output_dict["wind_delay_time"] / calculate_costs_input_dict["operational_hrs_per_day"]
        ) / calculate_costs_output_dict["time_construct_days"]
        # check if wind_delay_fraction is greater than 1, which would mean weather delays are longer than they can possibily be for the input data
        if wind_delay_fraction > 1:
            raise ValueError("{}: Error: Wind delay greater than 100%".format(type(self).__name__))
        calculate_costs_output_dict["wind_multiplier"] = 1 / (1 - wind_delay_fraction)

        # Calculating trenching cost:
        calculate_costs_output_dict["Days taken for trenching (equipment)"] = (
            calculate_costs_output_dict["trench_length_km"] / self._km_to_LF
        ) / calculate_costs_output_dict["trenching_equipment_daily_output"]
        calculate_costs_output_dict["Equipment cost of trenching per day {usd/day)"] = (
            calculate_costs_output_dict["trenching_cable_equipment_usd_per_hr"]
            * calculate_costs_input_dict["operational_hrs_per_day"]
        )
        calculate_costs_output_dict["Equipment Cost USD without weather delays"] = (
            calculate_costs_output_dict["Days taken for trenching (equipment)"]
            * calculate_costs_output_dict["Equipment cost of trenching per day {usd/day)"]
        )
        calculate_costs_output_dict["Equipment Cost USD with weather delays"] = (
            calculate_costs_output_dict["Equipment Cost USD without weather delays"]
            * calculate_costs_output_dict["wind_multiplier"]
        )

        if calculate_costs_input_dict["turbine_rating_MW"] >= 0.1:
            trenching_equipment_rental_cost_df = pd.DataFrame(
                [
                    [
                        "Equipment rental",
                        calculate_costs_output_dict["Equipment Cost USD with weather delays"],
                        "Collection",
                    ]
                ],
                columns=["Type of cost", "Cost USD", "Phase of construction"],
            )

        # switch for small DW
        else:
            if calculate_costs_output_dict["Equipment Cost USD with weather delays"] < 137:
                calculate_costs_output_dict["Equipment Cost USD with weather delays"] = 137  # cost of renting for a day
                trenching_equipment_rental_cost_df = pd.DataFrame(
                    [
                        [
                            "Equipment rental",
                            calculate_costs_output_dict["Equipment Cost USD with weather delays"],
                            "Collection",
                        ]
                    ],
                    columns=["Type of cost", "Cost USD", "Phase of construction"],
                )
            else:
                trenching_equipment_rental_cost_df = pd.DataFrame(
                    [
                        [
                            "Equipment rental",
                            calculate_costs_output_dict["Equipment Cost USD with weather delays"],
                            "Small DW Collection",
                        ]
                    ],
                    columns=["Type of cost", "Cost USD", "Phase of construction"],
                )

        # Calculating labor cost:
        calculate_costs_output_dict["Days taken for trenching (labor)"] = (
            calculate_costs_output_dict["trench_length_km"] / self._km_to_LF
        ) / calculate_costs_output_dict["trenching_labor_daily_output"]
        calculate_costs_output_dict["Labor cost of trenching per day (usd/day)"] = (
            calculate_costs_output_dict["trenching_labor_usd_per_hr"]
            * calculate_costs_input_dict["operational_hrs_per_day"]
            * calculate_costs_input_dict["overtime_multiplier"]
        )
        calculate_costs_output_dict["Total per diem costs (USD)"] = per_diem.sum()
        calculate_costs_output_dict["Labor Cost USD without weather delays"] = (
            calculate_costs_output_dict["Days taken for trenching (labor)"]
            * calculate_costs_output_dict["Labor cost of trenching per day (usd/day)"]
        ) + (
            calculate_costs_output_dict["Total per diem costs (USD)"]
            + calculate_costs_output_dict["managament_crew_cost_before_wind_delay"]
        )
        calculate_costs_output_dict["Labor Cost USD with weather delays"] = (
            calculate_costs_output_dict["Labor Cost USD without weather delays"]
            * calculate_costs_output_dict["wind_multiplier"]
        )

        if calculate_costs_input_dict["turbine_rating_MW"] >= 0.1:
            trenching_labor_cost_df = pd.DataFrame(
                [["Labor", calculate_costs_output_dict["Labor Cost USD with weather delays"], "Collection"]],
                columns=["Type of cost", "Cost USD", "Phase of construction"],
            )

        # switch for small DW
        else:
            trenching_labor_cost_df = pd.DataFrame(
                [["Labor", calculate_costs_output_dict["Labor Cost USD with weather delays"], "Small DW Collection"]],
                columns=["Type of cost", "Cost USD", "Phase of construction"],
            )

        # Calculate cable cost:
        cable_cost_usd_per_LF_df = pd.DataFrame(
            [["Materials", self._total_cable_cost, "Collection"]],
            columns=["Type of cost", "Cost USD", "Phase of construction"],
        )

        # Combine all calculated cost items into the 'collection_cost' dataframe:
        collection_cost = pd.DataFrame([], columns=["Type of cost", "Cost USD", "Phase of construction"])
        collection_cost = collection_cost.append(trenching_equipment_rental_cost_df)
        collection_cost = collection_cost.append(trenching_labor_cost_df)
        collection_cost = collection_cost.append(cable_cost_usd_per_LF_df)

        # Calculate Mobilization Cost and add to collection_cost dataframe.
        # For utility scale plants, mobilization is assumed to be 5% of the sum of labor, equipment, and material costs.
        # For distributed mode, mobilization is a calculated % that is a function of turbine size.
        if calculate_costs_input_dict["num_turbines"] > 10:
            calculate_costs_output_dict["mob_cost"] = collection_cost["Cost USD"].sum() * 0.05
        else:
            if calculate_costs_input_dict["turbine_rating_MW"] >= 0.1:
                calculate_costs_output_dict["mob_cost"] = collection_cost[
                    "Cost USD"
                ].sum() * self.mobilization_cost_multiplier(calculate_costs_input_dict["turbine_rating_MW"])

            # switch for small DW
            else:  # mobilization cost included in equipment rental cost
                calculate_costs_output_dict["mob_cost"] = 0.0

        mobilization_cost = pd.DataFrame(
            [["Mobilization", calculate_costs_output_dict["mob_cost"], "Collection"]],
            columns=["Type of cost", "Cost USD", "Phase of construction"],
        )
        collection_cost = collection_cost.append(mobilization_cost)

        calculate_costs_output_dict["total_collection_cost"] = collection_cost

        return collection_cost

    def outputs_for_detailed_tab(self, input_dict, output_dict):
        """
        Creates a list of dictionaries which can be used on their own or
        used to make a dataframe.

        Returns
        -------
        list(dict)
            A list of dicts, with each dict representing a row of the data.
        """
        result = []
        module = "Collection Cost"
        result.append(
            {
                "unit": "",
                "type": "variable",
                "variable_df_key_col_name": "Total Number of Turbines",
                "value": float(self.output_dict["total_turb"]),
            }
        )

        result.append(
            {
                "unit": "km",
                "type": "variable",
                "variable_df_key_col_name": "Total trench length",
                "value": float(self.output_dict["trench_length_km"]),
            }
        )

        result.append(
            {
                "unit": "km",
                "type": "variable",
                "variable_df_key_col_name": "Total cable length",
                "value": float(self.output_dict["total_cable_len_km"]),
            }
        )

        result.append(
            {
                "unit": "",
                "type": "variable",
                "variable_df_key_col_name": "Number of Turbines Per String in Full String",
                "value": float(self.output_dict["total_turb_per_string"]),
            }
        )
        result.append(
            {
                "unit": "",
                "type": "variable",
                "variable_df_key_col_name": "Number of Full Strings",
                "value": float(self.output_dict["num_full_strings"]),
            }
        )
        result.append(
            {
                "unit": "",
                "type": "variable",
                "variable_df_key_col_name": "Number of Turbines in Partial String",
                "value": float(self.output_dict["num_leftover_turb"]),
            }
        )
        result.append(
            {
                "unit": "",
                "type": "variable",
                "variable_df_key_col_name": "Number of Partial Strings",
                "value": float(self.output_dict["num_partial_strings"]),
            }
        )
        result.append(
            {
                "unit": "",
                "type": "variable",
                "variable_df_key_col_name": "Total number of strings full + partial",
                "value": float(self.output_dict["num_full_strings"] + self.output_dict["num_partial_strings"]),
            }
        )
        result.append(
            {
                "unit": "",
                "type": "variable",
                "variable_df_key_col_name": "Trench Length to Substation (km)",
                "value": float(self.output_dict["distance_to_grid_connection_km"]),
            }
        )
        result.append(
            {
                "unit": "",
                "type": "variable",
                "variable_df_key_col_name": "Cable Length to Substation (km)",
                "value": float(self.output_dict["cable_len_to_grid_connection_km"]),
            }
        )

        cables = ""
        n = 1  # to keep tab of number of cables input by user.
        for cable, specs in self.output_dict["cables"].items():
            if n == len(self.output_dict["cables"]):
                cables += str(cable)
            else:
                cables += str(cable) + "  ,  "

            for variable, value in specs.__dict__.items():
                if variable == "array_cable_len":
                    result.append(
                        {
                            "unit": "km",
                            "type": "variable",
                            "variable_df_key_col_name": "Array cable length for cable  " + cable,
                            "value": float(value),
                        }
                    )
                elif variable == "total_length":
                    result.append(
                        {
                            "unit": "km",
                            "type": "variable",
                            "variable_df_key_col_name": "Total cable length for cable  " + cable,
                            "value": float(value),
                        }
                    )

                elif variable == "total_cost":
                    result.append(
                        {
                            "unit": "usd",
                            "type": "variable",
                            "variable_df_key_col_name": "Total cable cost for cable  " + cable,
                            "value": float(value),
                        }
                    )
            n += 1

        result.append(
            {
                "unit": "",
                "type": "list",
                "variable_df_key_col_name": "Number of turbines per cable type in full strings [" + cables + "]",
                "value": str(self.output_dict["num_turb_per_cable"]),
            }
        )

        if self.input_dict["turbine_rating_MW"] > 0.1:
            for row in self.output_dict["management_crew"].itertuples():
                dashed_row = " <--> ".join(str(x) for x in list(row))
                result.append(
                    {
                        "unit": "",
                        "type": "dataframe",
                        "variable_df_key_col_name": "Labor type ID <--> Hourly rate USD per hour <--> Per diem USD per day <--> Operation <--> Crew type <--> Crew name <--> Number of workers <--> Per Diem Total <--> Hourly costs total <--> Crew total cost ",
                        "value": dashed_row,
                    }
                )

        result.append(
            {
                "unit": "",
                "type": "list",
                "variable_df_key_col_name": "Percent length of cable in partial string [" + cables + "]",
                "value": str(self.output_dict["perc_partial_string"]),
            }
        )

        for row in self.output_dict["total_collection_cost"].itertuples():
            dashed_row = "{} <--> {} <--> {}".format(row[1], row[3], math.ceil(row[2]))
            result.append(
                {
                    "unit": "",
                    "type": "dataframe",
                    "variable_df_key_col_name": "Type of Cost <--> Phase of Construction <--> Cost in USD ",
                    "value": dashed_row,
                    "last_number": row[2],
                }
            )

        for _dict in result:
            _dict["project_id_with_serial"] = self.project_name
            _dict["module"] = module

        self.output_dict["collection_cost_csv"] = result
        return result

    def run_module(self):
        """
        Runs the CollectionCost module and populates the IO dictionaries with calculated values.

        """

        try:
            self.create_ArraySystem()
            self.calculate_trench_properties(self.input_dict, self.output_dict)
            operation_data = self.estimate_construction_time(self.input_dict, self.output_dict)

            # pull only global inputs for weather delay from input_dict
            weather_data_keys = ("wind_shear_exponent", "weather_window")

            # specify collection-specific weather delay inputs
            self.weather_input_dict = dict(
                [(i, self.input_dict[i]) for i in self.input_dict if i in set(weather_data_keys)]
            )
            self.weather_input_dict[
                "start_delay_hours"
            ] = 0  # assume zero start for when collection construction begins (start at beginning of construction time)
            self.weather_input_dict["critical_wind_speed_m_per_s"] = self.input_dict[
                "critical_speed_non_erection_wind_delays_m_per_s"
            ]
            self.weather_input_dict["wind_height_of_interest_m"] = self.input_dict[
                "critical_height_non_erection_wind_delays_m"
            ]

            # Compute the duration of the construction for electrical collection
            duration_construction = operation_data["Time construct days"].max(skipna=True)
            days_per_month = 30
            duration_construction_months = duration_construction / days_per_month
            self.output_dict["collection_construction_months"] = duration_construction_months

            # compute and specify weather delay mission time for roads
            operational_hrs_per_day = self.input_dict["hour_day"][self.input_dict["time_construct"]]
            mission_time_hrs = duration_construction * operational_hrs_per_day
            self.weather_input_dict["mission_time_hours"] = int(mission_time_hrs)

            self.calculate_weather_delay(self.weather_input_dict, self.output_dict)
            self.calculate_costs(self.input_dict, self.output_dict)
            self.outputs_for_detailed_tab(self.input_dict, self.output_dict)
            self.output_dict["collection_cost_module_type_operation"] = self.outputs_for_costs_by_module_type_operation(
                input_df=self.output_dict["total_collection_cost"], project_id=self.project_name, total_or_turbine=True
            )
            return 0, 0  # module ran successfully
        except Exception as error:
            traceback.print_exc()
            print(f"Fail {self.project_name} CollectionCost")
            return 1, error  # module did not run successfully



"""Provides the `SparDesign` class (from OffshoreBOS)."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from numpy import exp, log
from wisdem.orbit.phases.design import DesignPhase


class SparDesign(DesignPhase):
    """Spar Substructure Design"""

    expected_config = {
        "site": {"depth": "m"},
        "plant": {"num_turbines": "int"},
        "turbine": {"turbine_rating": "MW"},
        "spar_design": {
            "stiffened_column_CR": "$/t (optional, default: 3120)",
            "tapered_column_CR": "$/t (optional, default: 4220)",
            "ballast_material_CR": "$/t (optional, default: 100)",
            "secondary_steel_CR": "$/t (optional, default: 7250)",
            "towing_speed": "km/h (optional, default: 6)",
        },
    }

    output_config = {
        "substructure": {
            "mass": "t",
            "ballasted_mass": "t",
            "unit_cost": "USD",
            "towing_speed": "km/h",
        }
    }

    def __init__(self, config, **kwargs):
        """
        Creates an instance of `SparDesign`.

        Parameters
        ----------
        config : dict
        """

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)
        self._design = self.config.get("spar_design", {})

        self._outputs = {}

    def run(self):
        """Main run function."""

        substructure = {
            "mass": self.unballasted_mass,
            "ballasted_mass": self.ballasted_mass,
            "unit_cost": self.substructure_cost,
            "towing_speed": self._design.get("towing_speed", 6),
        }

        self._outputs["substructure"] = substructure

    @property
    def stiffened_column_mass(self):
        """
        Calculates the mass of the stiffened column for a single spar in tonnes. From original OffshoreBOS model.
        """

        rating = self.config["turbine"]["turbine_rating"]
        depth = self.config["site"]["depth"]

        mass = 535.93 + 17.664 * rating ** 2 + 0.02328 * depth * log(depth)

        return mass

    @property
    def tapered_column_mass(self):
        """
        Calculates the mass of the atpered column for a single spar in tonnes. From original OffshoreBOS model.
        """

        rating = self.config["turbine"]["turbine_rating"]

        mass = 125.81 * log(rating) + 58.712

        return mass

    @property
    def stiffened_column_cost(self):
        """
        Calculates the cost of the stiffened column for a single spar. From original OffshoreBOS model.
        """

        cr = self._design.get("stiffened_column_CR", 3120)
        return self.stiffened_column_mass * cr

    @property
    def tapered_column_cost(self):
        """
        Calculates the cost of the tapered column for a single spar. From original OffshoreBOS model.
        """

        cr = self._design.get("tapered_column_CR", 4220)
        return self.tapered_column_mass * cr

    @property
    def ballast_mass(self):
        """
        Calculates the ballast mass of a single spar. From original OffshoreBOS model.
        """

        rating = self.config["turbine"]["turbine_rating"]
        mass = -16.536 * rating ** 2 + 1261.8 * rating - 1554.6

        return mass

    @property
    def ballast_cost(self):
        """
        Calculates the cost of ballast material for a single spar. From original OffshoreBOS model.
        """

        cr = self._design.get("ballast_material_CR", 100)
        return self.ballast_mass * cr

    @property
    def secondary_steel_mass(self):
        """
        Calculates the mass of the required secondary steel for a single
        spar. From original OffshoreBOS model.
        """

        rating = self.config["turbine"]["turbine_rating"]
        depth = self.config["site"]["depth"]

        mass = exp(3.58 + 0.196 * (rating ** 0.5) * log(rating) + 0.00001 * depth * log(depth))

        return mass

    @property
    def secondary_steel_cost(self):
        """
        Calculates the cost of the required secondary steel for a single
        spar. For original OffshoreBOS model.
        """

        cr = self._design.get("secondary_steel_CR", 7250)
        return self.secondary_steel_mass * cr

    @property
    def unballasted_mass(self):
        """Returns the unballasted mass of the spar substructure."""

        return self.stiffened_column_mass + self.tapered_column_mass + self.secondary_steel_mass

    @property
    def ballasted_mass(self):
        """Returns the ballasted mass of the spar substructure."""

        return self.unballasted_mass + self.ballast_mass

    @property
    def substructure_cost(self):
        """Returns the total cost (including ballast) of the spar substructure."""

        return self.stiffened_column_cost + self.tapered_column_cost + self.secondary_steel_cost + self.ballast_cost

    @property
    def detailed_output(self):
        """Returns detailed phase information."""

        _outputs = {
            "stiffened_column_mass": self.stiffened_column_mass,
            "stiffened_column_cost": self.stiffened_column_cost,
            "tapered_column_mass": self.tapered_column_mass,
            "tapered_column_cost": self.tapered_column_cost,
            "ballast_mass": self.ballast_mass,
            "ballast_cost": self.ballast_cost,
            "secondary_steel_mass": self.secondary_steel_mass,
            "secondary_steel_cost": self.secondary_steel_cost,
        }

        return _outputs

    @property
    def total_cost(self):
        """Returns total phase cost in $USD."""

        num = self.config["plant"]["num_turbines"]
        return num * self.substructure_cost

    @property
    def design_result(self):
        """Returns the result of `self.run()`"""

        if not self._outputs:
            raise Exception("Has `SparDesign` been ran yet?")

        return self._outputs
